import json

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_overrides_half_diamond(override, force):
    r"""
    pkgc -----> pkgb/0.1 --> pkga/0.1
       \--(override/force)-->pkga/0.2
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkga/0.2",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkgb")
    c.run("lock create pkgc")
    lock = json.loads(c.load("pkgc/conan.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.2" in requires
    assert "pkga/0.1" not in requires
    c.run("graph info pkgc --lockfile=pkgc/conan.lock --format=json")
    dependencies = json.loads(c.stdout)["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.2" in str(dependencies)
    assert "pkga/0.1" not in str(dependencies)
    # apply the lockfile to pkgb, should it lock to pkga/0.2
    c.run("graph info pkgb --lockfile=pkgc/conan.lock --format=json")
    dependencies = json.loads(c.stdout)["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.2" == dependencies["1"]["ref"]
    assert "pkga/0.1" == dependencies["1"]["require"]


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_overrides_half_diamond_ranges(override, force):
    r"""
       pkgc -----> pkgb/0.1 --> pkga/[>0.1 <0.2]
          \--(override/force)-->pkga/0.2
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/[>=0.1 <0.2]"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkga/0.2",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkgb")
    assert "pkga/0.2" not in c.out
    assert "pkga/0.1" in c.out
    c.run("lock create pkgc")
    lock = c.load("pkgc/conan.lock")
    assert "pkga/0.2" in lock
    assert "pkga/0.1" not in lock
    c.run("graph info pkgc --lockfile=pkgc/conan.lock")
    assert "pkga/0.2" in c.out
    assert "pkga/0.1" not in c.out


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_overrides_half_diamond_ranges_inverted(override, force):
    r""" the override is defining the lower bound of the range

       pkgc -----> pkgb/0.1 --> pkga/[>=0.1]
          \--(override/force)-->pkga/0.1
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/[>=0.1]"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkga/0.1",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkgb")
    assert "pkga/0.2" in c.out
    assert "pkga/0.1" not in c.out
    c.run("lock create pkgc")
    lock = c.load("pkgc/conan.lock")
    assert "pkga/0.1" in lock
    assert "pkga/0.2" not in lock
    c.run("graph info pkgc --lockfile=pkgc/conan.lock")
    assert "pkga/0.1" in c.out
    assert "pkga/0.2" not in c.out


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_overrides_diamond(override, force):
    r"""
    pkgd -----> pkgb/0.1 --> pkga/0.1
       \------> pkgc/0.1 --> pkga/0.2
       \--(override/force)-->pkga/0.3
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkga/0.2"),
            "pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkgc/0.1")
                                                            .with_requirement("pkga/0.3",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkga --version=0.3")
    c.run("create pkgb")
    c.run("create pkgc")
    c.run("lock create pkgd")
    lock = json.loads(c.load("pkgd/conan.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.3" in requires
    assert "pkga/0.2" not in requires
    assert "pkga/0.1" not in requires
    c.run("graph info pkgd --lockfile=pkgd/conan.lock --format=json")
    json_graph = json.loads(c.stdout)
    deps = json_graph["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.3" in str(deps)
    assert "pkga/0.2" not in str(deps)
    assert "pkga/0.1" not in str(deps)
    # Redundant assert, but checking "overrides" summary
    overrides = json_graph['graph']["overrides"]
    assert len(overrides) == 2
    assert overrides['pkga/0.1'] == ['pkga/0.3']
    assert overrides['pkga/0.2'] == ['pkga/0.3']

    # apply the lockfile to pkgb, should it lock to pkga/0.3
    c.run("graph info pkgb --lockfile=pkgd/conan.lock --format=json")
    json_graph = json.loads(c.stdout)
    deps = json_graph["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.3" == deps["1"]["ref"]
    assert "pkga/0.1" == deps["1"]["require"]
    # Redundant assert, but checking "overrides" summary
    overrides = json_graph['graph']["overrides"]
    assert len(overrides) == 1
    assert overrides["pkga/0.1"] == ["pkga/0.3"]


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_overrides_diamond_ranges(override, force):
    r"""
    pkgd -----> pkgb/0.1 --> pkga/[>=0.1 <0.2]
       \------> pkgc/0.1 --> pkga/[>=0.2 <0.3]
       \--(override/force)-->pkga/0.3
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/[>=0.1 <0.2]"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkga/[>=0.2 <0.3]"),
            "pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkgc/0.1")
                                                            .with_requirement("pkga/0.3",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkga --version=0.3")
    c.run("create pkgb")
    c.run("create pkgc")
    c.run("lock create pkgd")
    lock = json.loads(c.load("pkgd/conan.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.3" in requires
    assert "pkga/0.2" not in requires
    assert "pkga/0.1" not in requires
    c.run("graph info pkgd --lockfile=pkgd/conan.lock --format=json")
    dependencies = json.loads(c.stdout)["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.3" in str(dependencies)
    assert "pkga/0.2" not in str(dependencies)
    assert "pkga/0.1" not in str(dependencies)
    # apply the lockfile to pkgb, should it lock to pkga/0.3
    c.run("graph info pkgb --lockfile=pkgd/conan.lock --format=json")
    dependencies = json.loads(c.stdout)["graph"]["nodes"]["0"]["dependencies"]
    assert "pkga/0.3" in str(dependencies)
    assert "pkga/0.2" not in str(dependencies)
    assert "pkga/0.1" not in str(dependencies)


@pytest.mark.parametrize("override1, force1", [(True, False), (False, True)])
@pytest.mark.parametrize("override2, force2", [(True, False), (False, True)])
def test_overrides_multiple(override1, force1, override2, force2):
    r"""
    pkgd/0.1 -> pkgc/0.1 -> pkgb/0.1 -> pkga/0.1
      \           \--override---------> pkga/0.2
       \---override-------------------> pkga/0.3
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkga/0.2",
                                                                              override=override1,
                                                                              force=force1),
            "pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgc/0.1")
                                                            .with_requirement("pkga/0.3",
                                                                              override=override2,
                                                                              force=force2)
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkga --version=0.3")
    c.run("create pkgb")
    c.run("create pkgc --build=missing")
    c.run("lock create pkgd")
    lock = json.loads(c.load("pkgd/conan.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.3" in requires
    assert "pkga/0.2" not in requires
    assert "pkga/0.1" not in requires
    c.run("graph info pkgd --lockfile=pkgd/conan.lock")
    assert "pkga/0.3" in c.out
    assert "pkga/0.2#" not in c.out
    assert "pkga/0.1#" not in c.out  # appears in override information


def test_graph_different_overrides():
    r"""
    pkga -> toola/0.1 -> toolb/0.1 -> toolc/0.1
                \------override-----> toolc/0.2
    pkgb -> toola/0.2 -> toolb/0.2 -> toolc/0.1
                \------override-----> toolc/0.3
    pkgc -> toola/0.3 -> toolb/0.3 -> toolc/0.1
    """
    c = TestClient(light=True)
    c.save({"toolc/conanfile.py": GenConanfile("toolc"),
            "toolb/conanfile.py": GenConanfile("toolb").with_requires("toolc/0.1"),
            "toola/conanfile.py": GenConanfile("toola", "0.1").with_requirement("toolb/0.1")
                                                              .with_requirement("toolc/0.2",
                                                                                override=True),
            "toola2/conanfile.py": GenConanfile("toola", "0.2").with_requirement("toolb/0.2")
                                                               .with_requirement("toolc/0.3",
                                                                                 override=True),
            "toola3/conanfile.py": GenConanfile("toola", "0.3").with_requirement("toolb/0.3"),
            "pkga/conanfile.py": GenConanfile("pkga", "0.1").with_tool_requires("toola/0.1"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1")
                                                            .with_tool_requires("toola/0.2"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkgb/0.1")
                                                            .with_tool_requires("toola/0.3"),
            })
    c.run("create toolc --version=0.1")
    c.run("create toolc --version=0.2")
    c.run("create toolc --version=0.3")

    c.run("create toolb --version=0.1")
    c.run("create toolb --version=0.2")
    c.run("create toolb --version=0.3")

    c.run("create toola --build=missing")
    c.run("create toola2 --build=missing")
    c.run("create toola3 --build=missing")

    c.run("create pkga")
    c.run("create pkgb")
    c.run("lock create pkgc")
    lock = json.loads(c.load("pkgc/conan.lock"))
    requires = "\n".join(lock["build_requires"])
    assert "toolc/0.3" in requires
    assert "toolc/0.2" in requires
    assert "toolc/0.1" in requires

    c.run("graph info toolb --build-require --version=0.1 --lockfile=pkgc/conan.lock --format=json")
    # defaults to the non overriden
    c.assert_listed_require({"toolc/0.1": "Cache"}, build=True)
    # TODO: Solve it with build-order or manual overrides for the other packages


@pytest.mark.parametrize("override, force", [(True, False), (False, True)])
def test_introduced_conflict(override, force):
    """
    Using --lockfile-partial we can evaluate and introduce a new conflict
    pkgd -----> pkgb/[*] --> pkga/[>=0.1 <0.2]
    """
    c = TestClient(light=True)
    c.save({"pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb").with_requires("pkga/[>=0.1 <0.2]"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkga/[>=0.2 <0.3]"),
            "pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgb/[*]")
            })
    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkga --version=0.3")
    c.run("create pkgb --version=0.1")
    c.run("create pkgc")
    c.run("lock create pkgd")
    lock = json.loads(c.load("pkgd/conan.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.1" in requires
    assert "pkga/0.2" not in requires
    assert "pkga/0.3" not in requires
    # This will not be used thanks to the lockfile
    c.run("create pkgb --version=0.2")

    """
    This change in pkgd introduce a conflict
        Using --lockfile-partial we can evaluate and introduce a new conflict
        pkgd -----> pkgb/[*] --> pkga/[>=0.1 <0.2]
          |-------> pkgc/0.1 --> pkga/[>=0.2 <0.3]
    """
    c.save({"pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkgc/0.1")
            })

    c.run("graph info pkgd --lockfile=pkgd/conan.lock --lockfile-partial", assert_error=True)
    assert "Version conflict: Conflict between pkga/[>=0.2 <0.3] and pkga/0.1 in the graph" in c.out
    # Resolve the conflict with an override or force
    c.save({"pkgd/conanfile.py": GenConanfile("pkgd", "0.1").with_requirement("pkgb/0.1")
                                                            .with_requirement("pkgc/0.1")
                                                            .with_requirement("pkga/0.3",
                                                                              override=override,
                                                                              force=force)
            })
    c.run("graph info pkgd --lockfile=pkgd/conan.lock --lockfile-partial "
          "--lockfile-out=pkgd/conan2.lock --lockfile-clean")
    assert "pkgb/0.2" not in c.out
    assert "pkgb/0.1" in c.out
    lock = json.loads(c.load("pkgd/conan2.lock"))
    requires = "\n".join(lock["requires"])
    assert "pkga/0.3" in requires
    assert "pkga/0.1" not in requires
    assert "pkga/0.2" not in requires
    assert "pkgb/0.2" not in requires


def test_command_line_lockfile_overrides():
    """
    --lockfile-overrides cannot be abused to inject new overrides, only existing ones
    """
    c = TestClient(light=True)
    c.save({
            "pkga/conanfile.py": GenConanfile("pkga"),
            "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
            "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkgb/0.1"),
            })

    c.run("create pkga --version=0.1")
    c.run("create pkga --version=0.2")
    c.run("create pkgb")
    c.run('install pkgc --lockfile-overrides="{\'pkga/0.1\': [\'pkga/0.2\']}"', assert_error=True)
    assert "Cannot define overrides without a lockfile" in c.out
    c.run('lock create pkgc')
    c.run('install pkgc --lockfile-overrides="{\'pkga/0.1\': [\'pkga/0.2\']}"', assert_error=True)
    assert "Requirement 'pkga/0.2' not in lockfile" in c.out


def test_consecutive_installs():
    c = TestClient(light=True)
    c.save({
        "pkga/conanfile.py": GenConanfile("pkga"),
        "pkgb/conanfile.py": GenConanfile("pkgb", "0.1").with_requires("pkga/0.1"),
        "pkgc/conanfile.py": GenConanfile("pkgc", "0.1").with_requires("pkgb/0.1")
                                                        .with_requirement("pkga/0.2", override=True),
    })
    c.run("export pkga --version=0.1")
    c.run("export pkga --version=0.2")
    c.run("export pkgb")
    c.run("install pkgc --build=missing --lockfile-out=conan.lock")
    c.assert_overrides({"pkga/0.1": ["pkga/0.2"]})
    # This used to crash when overrides were not managed
    c.run("install pkgc --build=missing --lockfile=conan.lock --lockfile-out=conan.lock")
    c.assert_overrides({"pkga/0.1": ["pkga/0.2"]})



import textwrap

from parameterized import parameterized

from conan.internal.graph.graph_error import GraphProvidesError
from test.integration.graph.core.graph_manager_base import GraphManagerTest
from test.integration.graph.core.graph_manager_test import _check_transitive
from conan.test.utils.tools import GenConanfile, TestClient


class TestProvidesTest(GraphManagerTest):

    def test_direct_conflict(self):
        # app (provides feature)-> libb0.1 (provides feature)
        self.recipe_conanfile("libb/0.1", GenConanfile().with_provides("feature"))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_provides("feature").
                                           with_requires("libb/0.1"))
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 2 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[], dependents=[app])

    def test_transitive_conflict(self):
        # app (provides feature)-> libb0.1 -> libc/0.1 (provides feature)
        self.recipe_conanfile("libc/0.1", GenConanfile().with_provides("feature"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_requires("libc/0.1"))
        consumer = self.consumer_conanfile(GenConanfile("app", "0.1").with_provides("feature").
                                           with_requires("libb/0.1"))
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libb])
        self._check_node(libb, "libb/0.1#123", deps=[libc], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[], dependents=[libb])

    @parameterized.expand([(True,), (False,)])
    def test_branches_conflict(self, private):
        # app -> libb/0.1 (provides feature)
        #  \  -> libc/0.1 (provides feature)
        self.recipe_conanfile("libc/0.1", GenConanfile().with_provides("feature"))
        self.recipe_conanfile("libb/0.1", GenConanfile().with_provides("feature"))
        if private:
            consumer = self.consumer_conanfile(GenConanfile("app", "0.1").
                                               with_requirement("libb/0.1", visible=False).
                                               with_requirement("libc/0.1", visible=False))
        else:
            consumer = self.consumer_conanfile(GenConanfile("app", "0.1").
                                               with_requires("libb/0.1", "libc/0.1"))
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 3 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst

        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[], dependents=[app])

    def test_private_no_conflict(self):
        # app (provides libjpeg) -(private)-> br/v1 -(private)-> br_lib/v1(provides libjpeg)
        self.recipe_conanfile("br_lib/0.1", GenConanfile().with_provides("libjpeg"))
        self.recipe_conanfile("br/0.1", GenConanfile().with_requirement("br_lib/0.1", visible=False))
        path = self.consumer_conanfile(GenConanfile("app", "0.1").
                                       with_requirement("br/0.1", visible=False).
                                       with_provides("libjpeg"))

        deps_graph = self.build_consumer(path)
        assert 3 == len(deps_graph.nodes)

        app = deps_graph.root
        br = app.edges[0].dst
        br_lib = br.edges[0].dst

        self._check_node(app, "app/0.1", deps=[br])
        self._check_node(br, "br/0.1#123", deps=[br_lib], dependents=[app])
        self._check_node(br_lib, "br_lib/0.1#123", deps=[], dependents=[br])

        # node, include, link, build, run
        _check_transitive(app, [(br, True, True, False, False)])
        _check_transitive(br, [(br_lib, True, True, False, False)])

    def test_diamond_conflict(self):
        # app -> libb0.1 -> liba0.1
        #    \-> libc0.1 -> libd0.2 (provides liba)
        self.recipe_cache("liba/0.1")
        self.recipe_conanfile("libd/0.2", GenConanfile().with_provides("liba"))
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["libd/0.2"])

        consumer = self.recipe_consumer("app/0.1", ["libb/0.1", "libc/0.1"])
        deps_graph = self.build_consumer(consumer, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 5 == len(deps_graph.nodes)
        app = deps_graph.root
        libb = app.edges[0].dst
        libc = app.edges[1].dst
        liba1 = libb.edges[0].dst
        libd2 = libc.edges[0].dst
        self._check_node(app, "app/0.1", deps=[libb, libc])
        self._check_node(libb, "libb/0.1#123", deps=[liba1], dependents=[app])
        self._check_node(libc, "libc/0.1#123", deps=[libd2], dependents=[app])

        self._check_node(liba1, "liba/0.1#123", dependents=[libb])
        # TODO: Conflicted without revision
        self._check_node(libd2, "libd/0.2#123", dependents=[libc])

    def test_loop(self):
        # app -> libc0.1 -> libb0.1 -> liba0.1 ->|
        #             \<------(provides)---------|
        self.recipe_conanfile("liba/0.1", GenConanfile().with_provides("libc"))
        self.recipe_cache("libb/0.1", ["liba/0.1"])
        self.recipe_cache("libc/0.1", ["libb/0.1"])

        consumer = self.recipe_consumer("app/0.1", ["libc/0.1"])

        deps_graph = self.build_consumer(consumer, install=False)
        assert type(deps_graph.error) == GraphProvidesError

        assert 4 == len(deps_graph.nodes)

        app = deps_graph.root
        libc = app.edges[0].dst
        libb = libc.edges[0].dst
        liba = libb.edges[0].dst

        self._check_node(app, "app/0.1", deps=[libc])
        self._check_node(libc, "libc/0.1#123", deps=[libb], dependents=[app])
        self._check_node(libb, "libb/0.1#123", deps=[liba], dependents=[libc])
        self._check_node(liba, "liba/0.1#123", dependents=[libb])


class TestProvidesBuildRequire(GraphManagerTest):
    def test_build_require_no_conflict(self):
        # app (provides libjpeg) -(build)-> br/v1 -> br_lib/v1(provides libjpeg)
        self.recipe_conanfile("br_lib/0.1", GenConanfile().with_provides("libjpeg"))
        self.recipe_cache("br/0.1", ["br_lib/0.1"])
        path = self.consumer_conanfile(GenConanfile("app", "0.1").with_tool_requires("br/0.1").
                                       with_provides("libjpeg"))

        deps_graph = self.build_consumer(path)
        assert 3 == len(deps_graph.nodes)

        app = deps_graph.root
        br = app.edges[0].dst
        br_lib = br.edges[0].dst

        self._check_node(app, "app/0.1", deps=[br])
        self._check_node(br, "br/0.1#123", deps=[br_lib], dependents=[app])
        self._check_node(br_lib, "br_lib/0.1#123", deps=[], dependents=[br])

        # node, include, link, build, run
        _check_transitive(app, [(br, False, False, True, True)])
        _check_transitive(br, [(br_lib, True, True, False, False)])

    def test_transitive_br_no_conflict(self):
        # app (provides libjpeg) -> lib/v1 -(br)-> br/v1(provides libjpeg)
        self.recipe_conanfile("br/0.1", GenConanfile().with_provides("libjpeg"))
        self.recipe_conanfile("lib/0.1", GenConanfile().with_tool_requires("br/0.1"))
        path = self.consumer_conanfile(GenConanfile("app", "0.1").with_requires("lib/0.1").
                                       with_provides("libjpeg"))

        deps_graph = self.build_consumer(path)
        assert 3 == len(deps_graph.nodes)

        app = deps_graph.root
        lib = app.edges[0].dst
        br = lib.edges[0].dst

        self._check_node(app, "app/0.1", deps=[lib])
        self._check_node(lib, "lib/0.1#123", deps=[br], dependents=[app])
        self._check_node(br, "br/0.1#123", deps=[], dependents=[lib])

        # node, include, link, build, run
        _check_transitive(app, [(lib, True, True, False, False)])
        _check_transitive(lib, [(br, False, False, True, True)])

    def test_transitive_test_require_conflict(self):
        # app (provides libjpeg) -(test)-> br/v1 -> br_lib/v1(provides libjpeg)
        self.recipe_conanfile("br_lib/0.1", GenConanfile().with_provides("libjpeg"))
        self.recipe_cache("br/0.1", ["br_lib/0.1"])
        path = self.consumer_conanfile(GenConanfile("app", "0.1").with_test_requires("br/0.1").
                                       with_provides("libjpeg"))

        deps_graph = self.build_consumer(path, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 3 == len(deps_graph.nodes)

        app = deps_graph.root
        br = app.edges[0].dst
        br_lib = br.edges[0].dst

        self._check_node(app, "app/0.1", deps=[br])
        self._check_node(br, "br/0.1#123", deps=[br_lib], dependents=[app])
        self._check_node(br_lib, "br_lib/0.1#123", deps=[], dependents=[br])

    def test_two_br_conflict(self):
        # app -(build)-> br1/v1 (provides libjpeg)
        #   \ -(build)-> br2/v1 (provides libjpeg)
        self.recipe_conanfile("br1/0.1", GenConanfile().with_provides("libjpeg"))
        self.recipe_conanfile("br2/0.1", GenConanfile().with_provides("libjpeg"))
        path = self.consumer_conanfile(GenConanfile("app", "0.1")
                                       .with_tool_requires("br1/0.1", "br2/0.1"))
        deps_graph = self.build_consumer(path, install=False)

        assert type(deps_graph.error) == GraphProvidesError

        assert 3 == len(deps_graph.nodes)

        app = deps_graph.root
        br1 = app.edges[0].dst
        br2 = app.edges[1].dst

        self._check_node(app, "app/0.1", deps=[br1, br2])
        self._check_node(br1, "br1/0.1#123", deps=[], dependents=[app])
        self._check_node(br2, "br2/0.1#123", deps=[], dependents=[app])


def test_conditional():
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Recipe(ConanFile):
            requires = 'req/v1'
            options = {'conflict': [True, False]}
            default_options = {'conflict': False}

            def configure(self):
                if self.options.conflict:
                    self.provides = 'libjpeg'

            def package_info(self):
                self.info.clear()
    """)
    t = TestClient(light=True)
    t.save({'requires.py': GenConanfile("req", "v1").with_provides("libjpeg"),
            'app.py': conanfile})
    t.run("create requires.py")
    t.run("install app.py --name=app --version=version")
    t.run("install app.py --name=app --version=version -o app/*:conflict=True", assert_error=True)
    assert "ERROR: Provide Conflict: Both 'app/version' and 'req/v1' provide 'libjpeg'" in t.out


def test_self_build_require():
    c = TestClient()
    conanfile = textwrap.dedent("""\
        from conan import ConanFile
        from conan.tools.build import cross_building

        class Pkg(ConanFile):
            name = "grpc"
            version = "0.1"
            settings = "os"
            provides = "grpc-impl"
            def build_requirements(self):
                if cross_building(self):
                    self.tool_requires("grpc/0.1")
        """)
    c.save({'conanfile.py': conanfile})
    c.run("create . -s os=Windows -s:b os=Windows")
    c.assert_listed_binary({"grpc/0.1": ("ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715", "Build")})
    c.run("create . -s os=Linux -s:b os=Windows --build=missing")
    c.assert_listed_binary({"grpc/0.1": ("9a4eb3c8701508aa9458b1a73d0633783ecc2270", "Build")})
    c.assert_listed_binary({"grpc/0.1": ("ebec3dc6d7f6b907b3ada0c3d3cdc83613a2b715", "Cache")},
                           build=True)


def test_name_provide_error_message():
    tc = TestClient(light=True)
    tc.save({"libjepg/conanfile.py": GenConanfile("libjpeg", "0.1"),
             "mozjpeg/conanfile.py": GenConanfile("mozjpeg", "0.1").with_provides("libjpeg")})
    tc.run("create libjepg")
    tc.run("create mozjpeg")

    tc.run("graph info --requires=mozjpeg/0.1 --requires=libjpeg/0.1", assert_error=True)
    # This used to report that None was provided, but now it reports the name of the provides
    assert "ERROR: Provide Conflict: Both 'libjpeg/0.1' and 'mozjpeg/0.1' provide '['libjpeg']'" in tc.out



import os
import textwrap
import time

from conan.api.model import RecipeReference
from conan.test.utils.tools import TestClient, GenConanfile


class TestPyRequiresExtend:

    @staticmethod
    def _define_base(client):
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                def source(self):
                    self.output.info("My cool source!")
                def build(self):
                    self.output.info("My cool build!")
                def package(self):
                    self.output.info("My cool package!")
                def package_info(self):
                    self.output.info("My cool package_info!")
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")

    def test_reuse(self):
        client = TestClient(light=True, default_server_user=True)
        self._define_base(client)
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
            """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        package_id = client.created_package_id("pkg/0.1@user/testing")
        assert "pkg/0.1@user/testing: My cool source!" in client.out
        assert "pkg/0.1@user/testing: My cool build!" in client.out
        assert "pkg/0.1@user/testing: My cool package!" in client.out
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

        client.run("upload * --confirm -r default")
        client.run("remove * -c")
        client.run("install --requires=pkg/0.1@user/testing")
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out
        client.run("remove * -c")
        client.run("download pkg/0.1@user/testing#latest:* -r default")
        assert f"pkg/0.1@user/testing: Package installed {package_id}" in client.out
        # But it's broken with a single download
        client.run("install --requires=pkg/0.1@user/testing -nr", assert_error=True)
        assert "Cannot resolve python_requires 'base/1.1@user/testing'" in client.out
        # If pyrequires are expected, then first graph info -f=json and then get recipes from pkglist
        client.run("remove * -c")
        client.run("graph info --requires=pkg/0.1@user/testing -f=json",
                   redirect_stdout="graph_info.json")
        # We can even remove from the cache now (The pyrequires is already downloaded in above step)
        client.run("remove * -c")
        client.run("list --graph=graph_info.json --graph-recipes=* -f=json",
                   redirect_stdout="pkglist.json")
        client.run("download --list=pkglist.json -r default")
        assert "Downloading recipe 'base/1.1@user/testing" in client.out
        client.run("install --requires=pkg/0.1@user/testing -nr")
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

    def test_reuse_super(self):
        client = TestClient(light=True, default_server_user=True)
        self._define_base(client)
        reuse = textwrap.dedent("""
               from conan import ConanFile
               class PkgTest(ConanFile):
                   python_requires = "base/1.1@user/testing"
                   python_requires_extend = "base.MyConanfileBase"

                   def source(self):
                       super().source()
                       self.output.info("MY OWN SOURCE")
               """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("source . --name=pkg --version=0.1")
        assert "conanfile.py (pkg/0.1): My cool source!" in client.out
        assert "conanfile.py (pkg/0.1): MY OWN SOURCE" in client.out

    def test_reuse_dot(self):
        client = TestClient(light=True, default_server_user=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                def build(self):
                    self.output.info("My cool build!")
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=my.base --version=1.1 --user=user --channel=testing")
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "my.base/1.1@user/testing"
                python_requires_extend = "my.base.MyConanfileBase"
            """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: My cool build!" in client.out

    def test_with_alias(self):
        client = TestClient(light=True)
        self._define_base(client)
        client.alias("base/latest@user/testing", "base/1.1@user/testing")

        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                name = "pkg"
                version = "1.0"
                python_requires = "base/latest@user/testing"
            """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("install .", assert_error=True)
        assert "python-requires 'alias' are not supported in Conan 2.0" in client.out
        client.run("create .", assert_error=True)
        assert "python-requires 'alias' are not supported in Conan 2.0" in client.out

    def test_reuse_version_ranges(self):
        client = TestClient(light=True)
        self._define_base(client)

        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/[>1.0 <1.2]@user/testing"
                python_requires_extend = "base.MyConanfileBase"
            """)

        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        client.assert_listed_require({"base/1.1@user/testing": "Cache"}, python=True)
        assert "pkg/0.1@user/testing: My cool source!" in client.out
        assert "pkg/0.1@user/testing: My cool build!" in client.out
        assert "pkg/0.1@user/testing: My cool package!" in client.out
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

    def test_multiple_reuse(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class SourceBuild(ConanFile):
                def source(self):
                    self.output.info("My cool source!")
                def build(self):
                    self.output.info("My cool build!")
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=sourcebuild --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class PackageInfo(ConanFile):
                def package(self):
                    self.output.info("My cool package!")
                def package_info(self):
                    self.output.info("My cool package_info!")
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=packageinfo --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                python_requires = "sourcebuild/1.0@user/channel", "packageinfo/1.0@user/channel"
                python_requires_extend = "sourcebuild.SourceBuild", "packageinfo.PackageInfo"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: My cool source!" in client.out
        assert "pkg/0.1@user/testing: My cool build!" in client.out
        assert "pkg/0.1@user/testing: My cool package!" in client.out
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

    @staticmethod
    def test_transitive_access():
        client = TestClient(light=True)
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=base --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Helper(ConanFile):
                python_requires = "base/1.0@user/channel"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=helper --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                python_requires = "helper/1.0@user/channel"
                def build(self):
                    self.python_requires["base"]
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=channel")
        assert "pkg/0.1@user/channel: Created package" in client.out

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                python_requires = "helper/1.0@user/channel"
                python_requires_extend = "base.HelloConan"
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=channel")
        assert "pkg/0.1@user/channel: Created package" in client.out

    def test_multiple_requires_error(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            myvar = 123
            def myfunct():
                return 123
            class Pkg(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=pkg1 --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            myvar = 234
            def myfunct():
                return 234
            class Pkg(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=pkg2 --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                python_requires = "pkg1/1.0@user/channel", "pkg2/1.0@user/channel"
                def build(self):
                    self.output.info("PKG1 N: %s" % self.python_requires["pkg1"].conanfile.name)\
                               .info("PKG1 V: %s" % self.python_requires["pkg1"].conanfile.version)\
                               .info("PKG1 U: %s" % self.python_requires["pkg1"].conanfile.user)\
                               .info("PKG1 C: %s" % self.python_requires["pkg1"].conanfile.channel)\
                               .info("PKG1 : %s" % self.python_requires["pkg1"].module.myvar)\
                               .info("PKG2 : %s" % self.python_requires["pkg2"].module.myvar)\
                               .info("PKG1F : %s" % self.python_requires["pkg1"].module.myfunct())\
                               .info("PKG2F : %s" % self.python_requires["pkg2"].module.myfunct())
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=consumer --version=0.1 --user=user --channel=testing")
        assert "consumer/0.1@user/testing: PKG1 N: pkg1" in client.out
        assert "consumer/0.1@user/testing: PKG1 V: 1.0" in client.out
        assert "consumer/0.1@user/testing: PKG1 U: user" in client.out
        assert "consumer/0.1@user/testing: PKG1 C: channel" in client.out
        assert "consumer/0.1@user/testing: PKG1 : 123" in client.out
        assert "consumer/0.1@user/testing: PKG2 : 234" in client.out
        assert "consumer/0.1@user/testing: PKG1F : 123" in client.out
        assert "consumer/0.1@user/testing: PKG2F : 234" in client.out

    def test_local_import(self):
        client = TestClient(light=True, default_server_user=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            import mydata
            class MyConanfileBase(ConanFile):
                exports = "*.py"
                def source(self):
                    self.output.info(mydata.src)
                def build(self):
                    self.output.info(mydata.build)
                def package(self):
                    self.output.info(mydata.pkg)
                def package_info(self):
                    self.output.info(mydata.info)
            """)
        mydata = textwrap.dedent("""
            src = "My cool source!"
            build = "My cool build!"
            pkg = "My cool package!"
            info = "My cool package_info!"
            """)
        client.save({"conanfile.py": conanfile,
                     "mydata.py": mydata})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
            """)

        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        package_id = client.created_package_id("pkg/0.1@user/testing")
        assert "pkg/0.1@user/testing: My cool source!" in client.out
        assert "pkg/0.1@user/testing: My cool build!" in client.out
        assert "pkg/0.1@user/testing: My cool package!" in client.out
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

        client.run("upload * --confirm -r default")
        client.run("remove * -c")
        client.run("install --requires=pkg/0.1@user/testing")
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out
        client.run("remove * -c")
        client.run("download pkg/0.1@user/testing#*:* -r default")
        assert f"pkg/0.1@user/testing: Package installed {package_id}" in client.out

    def test_reuse_class_members(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                license = "MyLicense"
                author = "author@company.com"
                exports = "*.txt"
                exports_sources = "*.h"
                generators = "CMakeToolchain"
            """)
        client.save({"conanfile.py": conanfile,
                     "header.h": "some content"})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")

        reuse = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
                def build(self):
                    self.output.info("Exports sources! %s" % self.exports_sources)
                    self.output.info("HEADER CONTENT!: %s" % load(self, "header.h"))
                    self.output.info("License! %s" % self.license)
                    self.output.info("Author! %s" % self.author)
                    assert os.path.exists("conan_toolchain.cmake")
            """)
        client.save({"conanfile.py": reuse,
                     "header.h": "pkg new header contents",
                     "other.txt": "text"})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: Exports sources! *.h" in client.out
        assert "pkg/0.1@user/testing: Copied 1 '.txt' file: other.txt" in client.out
        assert "pkg/0.1@user/testing: Copied 1 '.h' file: header.h" in client.out
        assert "pkg/0.1@user/testing: License! MyLicense" in client.out
        assert "pkg/0.1@user/testing: Author! author@company.com" in client.out
        assert "pkg/0.1@user/testing: HEADER CONTENT!: pkg new header contents" in client.out
        ref = RecipeReference.loads("pkg/0.1@user/testing")
        assert os.path.exists(os.path.join(client.get_latest_ref_layout(ref).export(), "other.txt"))

    def test_reuse_system_requirements(self):
        # https://github.com/conan-io/conan/issues/7718
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           class MyConanfileBase(ConanFile):
               def system_requirements(self):
                   self.output.info("My system_requirements %s being called!" % self.name)
           """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
            """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: My system_requirements pkg being called!" in client.out

    def test_reuse_requirements(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
                   from conan import ConanFile
                   class MyConanfileBase(ConanFile):
                       def requirements(self):
                           self.output.info("My requirements %s being called!" % self.name)
                           self.requires("foo/1.0")
                   """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        client.save({"conanfile.py": GenConanfile("foo", "1.0")})
        client.run("create .")
        reuse = textwrap.dedent("""
                    from conan import ConanFile
                    class PkgTest(ConanFile):
                        python_requires = "base/1.1@user/testing"
                        python_requires_extend = "base.MyConanfileBase"
                    """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: My requirements pkg being called!" in client.out
        client.assert_listed_require({"foo/1.0#f5288356d9cc303f25cb05bccbad8fbb": "Cache"})

    def test_overwrite_class_members(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                license = "MyLicense"
                author = "author@company.com"
                settings = "os", # tuple!
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")

        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                license = "MIT"
                author = "frodo"
                settings = "arch", # tuple!
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"

                def init(self):
                    base = self.python_requires["base"].module.MyConanfileBase
                    self.settings = base.settings + self.settings
                    self.license = base.license

                def build(self):
                    self.output.info("License! %s" % self.license)
                    self.output.info("Author! %s" % self.author)
                    self.output.info("os: %s arch: %s" % (self.settings.get_safe("os"),
                                                          self.settings.arch))
            """)
        client.save({"conanfile.py": reuse})
        client.run("create . --name=pkg --version=0.1 -s os=Windows -s arch=armv7")
        assert "pkg/0.1: License! MyLicense" in client.out
        assert "pkg/0.1: Author! frodo" in client.out
        assert "pkg/0.1: os: Windows arch: armv7" in client.out

    def test_basic_option_inherit(self):
        client = TestClient()
        base = textwrap.dedent("""
            from conan import ConanFile
            class MyBase(object):
                settings = "os", "build_type", "arch"
                options = {"base_option": [1, 2, 3]}
                default_options = {"base_option": 2}

            class BaseConanFile(ConanFile):
                pass
            """)
        client.save({"conanfile.py": base})
        client.run("export . --name=base --version=1.0")
        derived = textwrap.dedent("""
            from conan import ConanFile
            class DerivedConan(ConanFile):
                settings = "os", "build_type", "arch"

                python_requires = "base/1.0"
                python_requires_extend = 'base.MyBase'

                def generate(self):
                    self.output.info(f"MYOPTION: {self.options.base_option}!!!!")
                """)
        client.save({"conanfile.py": derived})
        client.run("install .")
        assert "conanfile.py: MYOPTION: 2!!!!" in client.out

    def test_failure_init_method(self):
        client = TestClient()
        base = textwrap.dedent("""
            from conan import ConanFile
            class MyBase(object):
                settings = "os", "build_type", "arch"
                options = {"base_option": [True, False]}
                default_options = {"base_option": False}

            class BaseConanFile(ConanFile):
                pass
            """)
        client.save({"conanfile.py": base})
        client.run("export . --name=base --version=1.0")
        derived = textwrap.dedent("""
            from conan import ConanFile
            class DerivedConan(ConanFile):
                settings = "os", "build_type", "arch"

                python_requires = "base/1.0"
                python_requires_extend = 'base.MyBase'

                options = {"derived_option": [True, False]}
                default_options = {"derived_option": False}

                def init(self):
                    base = self.python_requires['base'].module.MyBase
                    self.options.update(base.options, base.default_options)
                """)
        client.save({"conanfile.py": derived})
        client.run("create . --name=pkg --version=0.1 -o base_option=True -o derived_option=True")
        assert "pkg/0.1: Created package" in client.out
        client.run("create . --name=pkg --version=0.1 -o whatever=True", assert_error=True)
        assert "Possible options are ['derived_option', 'base_option']" in client.out

    def test_transitive_imports_conflicts(self):
        # https://github.com/conan-io/conan/issues/3874
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            import myhelper
            class SourceBuild(ConanFile):
                exports = "*.py"
            """)
        helper = textwrap.dedent("""
            def myhelp(output):
                output.info("MyHelperOutput!")
            """)
        client.save({"conanfile.py": conanfile,
                     "myhelper.py": helper})
        client.run("export . --name=base1 --version=1.0 --user=user --channel=channel")
        client.save({"myhelper.py": helper.replace("MyHelperOutput!", "MyOtherHelperOutput!")})
        client.run("export . --name=base2 --version=1.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class MyConanfileBase(ConanFile):
                python_requires = "base2/1.0@user/channel", "base1/1.0@user/channel"
                def build(self):
                    self.python_requires["base1"].module.myhelper.myhelp(self.output)
                    self.python_requires["base2"].module.myhelper.myhelp(self.output)
            """)
        # This should work, even if there is a local "myhelper.py" file, which could be
        # accidentaly imported (and it was, it was a bug)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: MyHelperOutput!" in client.out
        assert "pkg/0.1@user/testing: MyOtherHelperOutput!" in client.out

        # Now, the same, but with "clean_first=True", should keep working
        client.save({"conanfile.py": conanfile}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: MyHelperOutput!" in client.out
        assert "pkg/0.1@user/testing: MyOtherHelperOutput!" in client.out

    def test_update(self):
        client = TestClient(light=True, default_server_user=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            somevar = 42
            class MyConanfileBase(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        client.run("upload * --confirm -r default")

        client2 = TestClient(light=True, servers=client.servers, inputs=["user", "password"])
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
                def configure(self):
                    self.output.info("PYTHON REQUIRE VAR %s"
                                     % self.python_requires["base"].module.somevar)
        """)

        client2.save({"conanfile.py": reuse})
        client2.run("install .")
        assert "conanfile.py: PYTHON REQUIRE VAR 42" in client2.out

        client.save({"conanfile.py": conanfile.replace("42", "143")})
        time.sleep(1)  # guarantee time offset
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        client.run("upload * --confirm -r default")

        client2.run("install . --update")
        assert "conanfile.py: PYTHON REQUIRE VAR 143" in client2.out

    def test_update_ranges(self):
        # Same as the above, but using a version range, and no --update
        # https://github.com/conan-io/conan/issues/4650#issuecomment-497464305
        client = TestClient(light=True, default_server_user=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            somevar = 42
            class MyConanfileBase(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=base --version=1.1 --user=user --channel=testing")
        client.run("upload * --confirm -r default")

        client2 = TestClient(light=True, servers=client.servers, inputs=["user", "password"])
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/[>1.0]@user/testing"
                python_requires_extend = "base.MyConanfileBase"
                def configure(self):
                    self.output.info("PYTHON REQUIRE VAR %s"
                                     % self.python_requires["base"].module.somevar)
        """)

        client2.save({"conanfile.py": reuse})
        client2.run("install .")
        assert "conanfile.py: PYTHON REQUIRE VAR 42" in client2.out

        client.save({"conanfile.py": conanfile.replace("42", "143")})
        # Make sure to bump the version!
        client.run("export . --name=base --version=1.2 --user=user --channel=testing")
        client.run("upload * --confirm -r default")

        client2.run("install . --update")
        assert "conanfile.py: PYTHON REQUIRE VAR 143" in client2.out

    def test_duplicate_pyreq(self):
        t = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class PyReq(ConanFile):
                pass
        """)
        t.save({"conanfile.py": conanfile})
        t.run("export . --name=pyreq --version=1.0 --user=user --channel=channel")
        t.run("export . --name=pyreq --version=2.0 --user=user --channel=channel")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Lib(ConanFile):
                python_requires = "pyreq/1.0@user/channel", "pyreq/2.0@user/channel"
        """)
        t.save({"conanfile.py": conanfile})
        t.run("create . --name=name --version=version --user=user --channel=channel",
              assert_error=True)
        assert "ERROR: Error loading conanfile" in t.out
        assert "The python_require 'pyreq' already exists" in t.out

    def test_local_build(self):
        client = TestClient(light=True)
        client.save({"conanfile.py": "var=42\n" + str(GenConanfile())})
        client.run("export . --name=tool --version=0.1 --user=user --channel=channel")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                python_requires = "tool/0.1@user/channel"
                def source(self):
                    self.output.info("Pkg1 source: %s" % self.python_requires["tool"].module.var)
                def build(self):
                    self.output.info("Pkg1 build: %s" % self.python_requires["tool"].module.var)
                def package(self):
                    self.output.info("Pkg1 package: %s" % self.python_requires["tool"].module.var)
            """)
        client.save({"conanfile.py": conanfile})
        client.run("source .")
        assert "conanfile.py: Pkg1 source: 42" in client.out
        client.run("install .")
        client.run("build .")
        assert "conanfile.py: Pkg1 build: 42" in client.out
        client.run("export-pkg . --name=pkg1 --version=0.1 --user=user --channel=testing")

    def test_reuse_name_version(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os

            class Source(object):
                def set_name(self):
                    self.name = load(self, "name.txt")

                def set_version(self):
                    self.version = load(self, "version.txt")

            class MyConanfileBase(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=tool --version=0.1 --user=user --channel=channel")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                python_requires = "tool/0.1@user/channel"
                python_requires_extend = "tool.Source"
                def source(self):
                    self.output.info("Pkg1 source: %s:%s" % (self.name, self.version))
                def build(self):
                    self.output.info("Pkg1 build: %s:%s" % (self.name, self.version))
                def package(self):
                    self.output.info("Pkg1 package: %s:%s" % (self.name, self.version))
            """)
        client.save({"conanfile.py": conanfile,
                     "name.txt": "mypkg",
                     "version.txt": "myversion"})
        client.run("export .")
        assert "mypkg/myversion: Exported" in client.out
        client.run("create .")
        assert "mypkg/myversion: Pkg1 source: mypkg:myversion" in client.out
        assert "mypkg/myversion: Pkg1 build: mypkg:myversion" in client.out
        assert "mypkg/myversion: Pkg1 package: mypkg:myversion" in client.out

    def test_reuse_name_version_override(self):
        client = TestClient(light=True)

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os

            class Source(object):
                def set_name(self):
                    self.name = load(self, "name.txt")

                def set_version(self):
                    self.version = load(self, "version.txt")

            class MyConanfileBase(ConanFile):
                pass
            """)
        client.save({"conanfile.py": conanfile})
        client.run("export . --name=tool --version=0.1 --user=user --channel=channel")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                python_requires = "tool/0.1@user/channel"
                python_requires_extend = "tool.Source"

                def set_version(self):
                    super().set_version()
                    self.version = self.version + "2"

                def source(self):
                    self.output.info("Pkg1 source: %s:%s" % (self.name, self.version))
                def build(self):
                    self.output.info("Pkg1 build: %s:%s" % (self.name, self.version))
                def package(self):
                    self.output.info("Pkg1 package: %s:%s" % (self.name, self.version))
            """)
        client.save({"conanfile.py": conanfile,
                     "name.txt": "mypkg",
                     "version.txt": "myversion"})
        client.run("export .")
        assert "mypkg/myversion2: Exported" in client.out
        client.run("create .")
        assert "mypkg/myversion2: Pkg1 source: mypkg:myversion2" in client.out
        assert "mypkg/myversion2: Pkg1 build: mypkg:myversion2" in client.out
        assert "mypkg/myversion2: Pkg1 package: mypkg:myversion2" in client.out

    def test_reuse_export_sources(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                exports = "*"
            """)
        client.save({"conanfile.py": conanfile,
                     "file.h": "myheader",
                     "folder/other.h": "otherheader"})
        client.run("export . --name=tool --version=0.1 --user=user --channel=channel")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os
            class MyConanfileBase(ConanFile):
                python_requires = "tool/0.1@user/channel"
                def source(self):
                    sources = self.python_requires["tool"].path
                    file_h = os.path.join(sources, "file.h")
                    other_h = os.path.join(sources, "folder/other.h")
                    self.output.info("Source: tool header: %s" % load(self, file_h))
                    self.output.info("Source: tool other: %s" % load(self, other_h))
                def build(self):
                    sources = self.python_requires["tool"].path
                    file_h = os.path.join(sources, "file.h")
                    other_h = os.path.join(sources, "folder/other.h")
                    self.output.info("Build: tool header: %s" % load(self, file_h))
                    self.output.info("Build: tool other: %s" % load(self, other_h))
                def package(self):
                    sources = self.python_requires["tool"].path
                    file_h = os.path.join(sources, "file.h")
                    other_h = os.path.join(sources, "folder/other.h")
                    self.output.info("Package: tool header: %s" % load(self, file_h))
                    self.output.info("Package: tool other: %s" % load(self, other_h))
            """)
        client.save({"conanfile.py": conanfile,
                     "name.txt": "MyPkg",
                     "version.txt": "MyVersion"})
        client.run("export . --name=pkg --version=1.0 --user=user --channel=channel")
        assert "pkg/1.0@user/channel: Exported" in client.out
        client.run("create . --name=pkg --version=1.0 --user=user --channel=channel")
        assert "pkg/1.0@user/channel: Source: tool header: myheader" in client.out
        assert "pkg/1.0@user/channel: Source: tool other: otherheader" in client.out
        assert "pkg/1.0@user/channel: Build: tool header: myheader" in client.out
        assert "pkg/1.0@user/channel: Build: tool other: otherheader" in client.out
        assert "pkg/1.0@user/channel: Package: tool header: myheader" in client.out
        assert "pkg/1.0@user/channel: Package: tool other: otherheader" in client.out

        # The local flow
        client.run("install .")
        client.run("source .")
        assert "conanfile.py: Source: tool header: myheader" in client.out
        assert "conanfile.py: Source: tool other: otherheader" in client.out
        client.run("build .")
        assert "conanfile.py: Build: tool header: myheader" in client.out
        assert "conanfile.py: Build: tool other: otherheader" in client.out

    def test_reuse_editable_exports(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class MyConanfileBase(ConanFile):
                exports = "*"
            """)
        client.save({"conanfile.py": conanfile,
                     "file.h": "myheader",
                     "folder/other.h": "otherheader"})
        client.run("editable add . --name=tool --version=0.1")

        conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.files import load
            import os
            class MyConanfileBase(ConanFile):
                python_requires = "tool/0.1"
                def source(self):
                    sources = self.python_requires["tool"].path
                    file_h = os.path.join(sources, "file.h")
                    other_h = os.path.join(sources, "folder/other.h")
                    self.output.info("Source: tool header: %s" % load(self, file_h))
                    self.output.info("Source: tool other: %s" % load(self, other_h))
                def build(self):
                    sources = self.python_requires["tool"].path
                    file_h = os.path.join(sources, "file.h")
                    other_h = os.path.join(sources, "folder/other.h")
                    self.output.info("Build: tool header: %s" % load(self, file_h))
                    self.output.info("Build: tool other: %s" % load(self, other_h))
            """)

        client2 = TestClient(light=True, cache_folder=client.cache_folder)
        client2.save({"conanfile.py": conanfile,
                      "name.txt": "MyPkg",
                      "version.txt": "MyVersion"})

        # The local flow
        client2.run("install .")
        client2.run("source .")
        assert "conanfile.py: Source: tool header: myheader" in client2.out
        assert "conanfile.py: Source: tool other: otherheader" in client2.out
        client2.run("build .")
        assert "conanfile.py: Build: tool header: myheader" in client2.out
        assert "conanfile.py: Build: tool other: otherheader" in client2.out

    def test_build_id(self):
        client = TestClient(light=True, default_server_user=True)
        self._define_base(client)
        reuse = textwrap.dedent("""
            from conan import ConanFile
            class PkgTest(ConanFile):
                python_requires = "base/1.1@user/testing"
                python_requires_extend = "base.MyConanfileBase"
                def build_id(self):
                    pass
            """)
        client.save({"conanfile.py": reuse}, clean_first=True)
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: My cool source!" in client.out
        assert "pkg/0.1@user/testing: My cool build!" in client.out
        assert "pkg/0.1@user/testing: My cool package!" in client.out
        assert "pkg/0.1@user/testing: My cool package_info!" in client.out

    def test_options_errors(self):
        c = TestClient(light=True)
        base = textwrap.dedent("""
            from conan import ConanFile
            class BaseConan:
                options = {"base": [True, False]}
                default_options = {"base": True}

            class PyReq(ConanFile):
                name = "base"
                version = "1.0.0"
                package_type = "python-require"
                """)
        derived = textwrap.dedent("""
            import conan

            class DerivedConan(conan.ConanFile):
                name = "derived"
                python_requires = "base/1.0.0"
                python_requires_extend = "base.BaseConan"
                options = {"derived": [True, False]}
                default_options = {"derived": False}

                def init(self):
                    base = self.python_requires["base"].module.BaseConan
                    self.options.update(base.options, base.default_options)

                def generate(self):
                    self.output.info(f"OptionBASE: {self.options.base}")
                    self.output.info(f"OptionDERIVED: {self.options.derived}")
            """)
        c.save({"base/conanfile.py": base,
                "derived/conanfile.py": derived})
        c.run("create base")
        c.run("install derived")
        assert "OptionBASE: True" in c.out
        assert "OptionDERIVED: False" in c.out

    def test_options_default_update(self):
        c = TestClient(light=True)
        base = textwrap.dedent("""
            from conan import ConanFile
            class BaseConan:
                options = {"base": [True, False]}
                default_options = {"base": True}

            class PyReq(ConanFile):
                name = "base"
                version = "1.0.0"
                package_type = "python-require"
                """)
        derived = textwrap.dedent("""
            import conan

            class DerivedConan(conan.ConanFile):
                name = "derived"
                python_requires = "base/1.0.0"
                python_requires_extend = "base.BaseConan"
                options = {"derived": [True, False]}
                default_options = {"derived": False}

                def init(self):
                    base = self.python_requires["base"].module.BaseConan
                    self.options.update(base.options, base.default_options)
                    self.options.base = False

                def generate(self):
                    self.output.info(f"OptionBASE: {self.options.base}")
                    self.output.info(f"OptionDERIVED: {self.options.derived}")
            """)
        c.save({"base/conanfile.py": base,
                "derived/conanfile.py": derived})
        c.run("create base")
        c.run("install derived")
        assert "OptionBASE: False" in c.out
        assert "OptionDERIVED: False" in c.out

        c.run("install derived -o=&:base=True")
        assert "OptionBASE: True" in c.out
        assert "OptionDERIVED: False" in c.out


def test_transitive_python_requires():
    # https://github.com/conan-io/conan/issues/8546
    client = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        myvar = 123
        def myfunct():
            return 234
        class SharedFunction(ConanFile):
            name = "shared-function"
            version = "1.0"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class BaseClass(ConanFile):
            name = "base-class"
            version = "1.0"
            python_requires = "shared-function/1.0@user/channel"
            def build(self):
                pyreqs = self.python_requires
                v = pyreqs["shared-function"].module.myvar  # v will be 123
                f = pyreqs["shared-function"].module.myfunct()  # f will be 234
                self.output.info("%s, %s" % (v, f))
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            name = "consumer"
            version = "1.0"
            python_requires = "base-class/1.0@user/channel"
            python_requires_extend = "base-class.BaseClass"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --user=user --channel=channel")
    assert "consumer/1.0@user/channel: Calling build()\nconsumer/1.0@user/channel: 123, 234" in \
           client.out


def test_transitive_diamond_python_requires():
    client = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        myvar = 123
        def myfunct():
            return 234
        class SharedFunction(ConanFile):
            name = "shared-function"
            version = "1.0"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        myvar = 222
        def myfunct():
            return 2222
        class SharedFunction2(ConanFile):
            name = "shared-function2"
            version = "1.0"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class BaseClass(ConanFile):
            name = "base-class"
            version = "1.0"
            python_requires = "shared-function/1.0@user/channel", "shared-function2/1.0@user/channel"
            def build(self):
                pyreqs = self.python_requires
                v = pyreqs["shared-function"].module.myvar  # v will be 123
                f = pyreqs["shared-function2"].module.myfunct()  # f will be 2222
                self.output.info("%s, %s" % (v, f))
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class BaseClass2(ConanFile):
            name = "base-class2"
            version = "1.0"
            python_requires = "shared-function/1.0@user/channel", "shared-function2/1.0@user/channel"
            def package(self):
                pyreqs = self.python_requires
                v = pyreqs["shared-function2"].module.myvar  # v will be 222
                f = pyreqs["shared-function"].module.myfunct()  # f will be 234
                self.output.info("%s, %s" % (v, f))
        """)
    client.save({"conanfile.py": conanfile})
    client.run("export . --user=user --channel=channel")

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            name = "consumer"
            version = "1.0"
            python_requires = "base-class/1.0@user/channel", "base-class2/1.0@user/channel"
            python_requires_extend = "base-class.BaseClass", "base-class2.BaseClass2"
        """)
    client.save({"conanfile.py": conanfile})
    client.run("create . --user=user --channel=channel")
    assert "consumer/1.0@user/channel: Calling build()\nconsumer/1.0@user/channel: 123, 2222" in \
           client.out
    assert "consumer/1.0@user/channel: Calling package()\nconsumer/1.0@user/channel: 222, 234" in \
           client.out


class TestConflictPyRequires:
    # https://github.com/conan-io/conan/issues/15016
    def test_diamond_conflict_fixed(self):
        c = TestClient(light=True)

        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "sub1/conanfile.py": GenConanfile("sub1", "1.0").with_python_requires("tool/1.0"),
                "sub2/conanfile.py": GenConanfile("sub2", "1.0").with_python_requires("tool/1.1"),
                "app/conanfile.py": GenConanfile().with_python_requires("sub1/1.0", "sub2/1.0")})
        c.run("export tool --version=1.0")
        c.run("export tool --version=1.1")
        c.run("export sub1")
        c.run("export sub2")
        c.run("install app", assert_error=True)
        assert "Conflict in py_requires tool/1.0 - tool/1.1" in c.out

    def test_diamond_conflict_ranges(self):
        c = TestClient(light=True)

        c.save({"tool/conanfile.py": GenConanfile("tool"),
                "sub1/conanfile.py": GenConanfile("sub1", "1.0").with_python_requires("tool/[*]"),
                "sub2/conanfile.py": GenConanfile("sub2", "1.0").with_python_requires("tool/1.0"),
                "app/conanfile.py": GenConanfile().with_python_requires("sub1/1.0", "sub2/1.0")})
        c.run("export tool --version=1.0")
        c.run("export tool --version=1.1")
        c.run("export sub1")
        c.run("export sub2")
        c.run("install app", assert_error=True)
        assert "Conflict in py_requires tool/1.1 - tool/1.0" in c.out


def test_multiple_reuse():
    """ test how to enable the multiple code reuse for custom user generators
        # https://github.com/conan-io/conan/issues/11589
    """

    c = TestClient(light=True)
    common = textwrap.dedent("""
        from conan import ConanFile
        def mycommon():
            return 42
        class Common(ConanFile):
            name = "common"
            version = "0.1"
        """)
    tool = textwrap.dedent("""
        from conan import ConanFile

        class MyGenerator:
            common = None
            def __init__(self, conanfile):
                self.conanfile = conanfile
            def generate(self):
                self.conanfile.output.info("VALUE TOOL: {}!!!".format(MyGenerator.common.mycommon()))

        class Tool(ConanFile):
            name = "tool"
            version = "0.1"
            python_requires = "common/0.1"
            def init(self):
                MyGenerator.common = self.python_requires["common"].module
        """)
    consumer = textwrap.dedent("""
        from conan import ConanFile
        class Consumer(ConanFile):
            python_requires = "tool/0.1", "common/0.1"
            def generate(self):
                mycommon = self.python_requires["common"].module.mycommon
                self.output.info("VALUE COMMON: {}!!!".format(mycommon()))
                mygenerator = self.python_requires["tool"].module.MyGenerator(self)
                mygenerator.generate()
        """)
    c.save({"common/conanfile.py": common,
            "tool/conanfile.py": tool,
            "consumer/conanfile.py": consumer})
    c.run("export common")
    c.run("export tool")
    c.run("install consumer")
    assert "VALUE COMMON: 42!!!" in c.out
    assert "VALUE TOOL: 42!!!" in c.out


class TestTestPackagePythonRequire:
    def test_test_package_python_requires(self):
        """ test how to test_package a python_require
        """

        c = TestClient(light=True)
        c.save({"conanfile.py": GenConanfile("mytool", "0.1")})
        c.run("create .")
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            def mycommon():
                return 42
            class Common(ConanFile):
                name = "common"
                version = "0.1"
                package_type = "python-require"
            """)
        test = textwrap.dedent("""
            from conan import ConanFile

            class Tool(ConanFile):
                python_requires = "tested_reference_str"
                def test(self):
                    self.output.info("{}!!!".format(self.python_requires["common"].module.mycommon()))
            """)
        c.save({"conanfile.py": conanfile,
                "test_package/conanfile.py": test})
        c.run("create .")
        assert "WARN: deprecated: test_package/conanfile.py" not in c.out
        assert "common/0.1 (test package): 42!!!" in c.out

        c.run("test test_package common/0.1")
        assert "WARN: deprecated: test_package/conanfile.py" not in c.out
        assert "common/0.1 (test package): 42!!!" in c.out

        # https://github.com/conan-io/conan/issues/18224
        c.save({"myprofile": "[tool_requires]\nmytool/0.1"})
        c.run("create . -pr=myprofile")
        assert "common/0.1 (test package): 42!!!" in c.out

    def test_test_package_python_requires_configs(self):
        """ test how to test_package a python_require with various configurations
        """

        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            def mycommon(build_type):
                return str(build_type).upper() + "OK"
            class Common(ConanFile):
                name = "common"
                version = "0.1"
                package_type = "python-require"
            """)
        test = textwrap.dedent("""
            from conan import ConanFile

            class Tool(ConanFile):
                settings = "build_type"
                def test(self):
                    result = self.python_requires["common"].module.mycommon(self.settings.build_type)
                    self.output.info("{}!!!".format(result))
            """)
        c.save({"conanfile.py": conanfile,
                "test_package/conanfile.py": test})
        c.run("create . ")
        assert "common/0.1 (test package): RELEASEOK!!!" in c.out
        assert "WARN: deprecated: test_package/conanfile.py should declare 'python_requires" in c.out
        c.run("create . -s build_type=Debug")
        assert "common/0.1 (test package): DEBUGOK!!!" in c.out


class TestResolveRemote:
    def test_resolve_remote_export(self):
        """ a "conan export" command should work even when a python_requires
        is in the server
        """
        c = TestClient(light=True, default_server_user=True)
        c.save({"common/conanfile.py": GenConanfile("tool", "0.1"),
                "pkg/conanfile.py": GenConanfile("pkg", "0.1").with_python_requires("tool/0.1")})
        c.run("export common")
        c.run("upload * -r=default -c")
        c.run("remove * -c")
        c.run("create pkg")
        assert "tool/0.1: Downloaded recipe" in c.out

        c.run("remove * -c")
        c.run("export pkg")
        assert "tool/0.1: Downloaded recipe" in c.out

        c.run("remove * -c")
        c.run("export pkg -r=default")
        assert "tool/0.1: Downloaded recipe" in c.out

        c.run("remove * -c")
        c.run("create pkg -r=default")
        assert "tool/0.1: Downloaded recipe" in c.out

    def test_missing_python_require_error(self):
        """ make sure the error is clear enough for users UX
        """
        c = TestClient(light=True)
        c.save({"pkg/conanfile.py": GenConanfile("pkg", "0.1").with_python_requires("tool/0.1")})
        c.run("create pkg", assert_error=True)
        assert "Cannot resolve python_requires 'tool/0.1'" in c.out


class TestTransitiveExtend:
    # https://github.com/conan-io/conan/issues/10511
    # https://github.com/conan-io/conan/issues/10565
    def test_transitive_extend(self):
        client = TestClient(light=True)
        company = textwrap.dedent("""
            from conan import ConanFile
            class CompanyConanFile(ConanFile):
                name = "company"
                version = "1.0"

                def msg1(self):
                    return "company"
                def msg2(self):
                    return "company"
            """)
        project = textwrap.dedent("""
            from conan import ConanFile
            class ProjectBaseConanFile(ConanFile):
                name = "project"
                version = "1.0"

                python_requires = "company/1.0"
                python_requires_extend = "company.CompanyConanFile"

                def msg1(self):
                    return "project"
            """)
        consumer = textwrap.dedent("""
            from conan import ConanFile
            class Base(ConanFile):
                name = "consumer"
                version = "1.0"
                python_requires = "project/1.0"
                python_requires_extend = "project.ProjectBaseConanFile"
                def generate(self):
                    self.output.info("Msg1:{}!!!".format(self.msg1()))
                    self.output.info("Msg2:{}!!!".format(self.msg2()))
                """)
        client.save({"company/conanfile.py": company,
                     "project/conanfile.py": project,
                     "consumer/conanfile.py": consumer})
        client.run("export company")
        client.run("export project")
        client.run("install consumer")
        assert "conanfile.py (consumer/1.0): Msg1:project!!!" in client.out
        assert "conanfile.py (consumer/1.0): Msg2:company!!!" in client.out

    def test_transitive_extend2(self):
        client = TestClient(light=True)
        company = textwrap.dedent("""
            from conan import ConanFile
            class CompanyConanFile(ConanFile):
                name = "company"
                version = "1.0"

            class CompanyBase:
                def msg1(self):
                    return "company"
                def msg2(self):
                    return "company"
            """)
        project = textwrap.dedent("""
            from conan import ConanFile
            class ProjectBase:
                def msg1(self):
                    return "project"

            class ProjectBaseConanFile(ConanFile):
                name = "project"
                version = "1.0"
                python_requires = "company/1.0"

                def init(self):
                    pkg_name, base_class_name = "company", "CompanyBase"
                    base_class = getattr(self.python_requires[pkg_name].module, base_class_name)
                    global ProjectBase
                    ProjectBase = type('ProjectBase', (ProjectBase, base_class, object), {})
            """)
        consumer = textwrap.dedent("""
            from conan import ConanFile
            class Base(ConanFile):
                name = "consumer"
                version = "1.0"
                python_requires = "project/1.0"
                python_requires_extend = "project.ProjectBase"
                def generate(self):
                    self.output.info("Msg1:{}!!!".format(self.msg1()))
                    self.output.info("Msg2:{}!!!".format(self.msg2()))
                """)
        client.save({"company/conanfile.py": company,
                     "project/conanfile.py": project,
                     "consumer/conanfile.py": consumer})
        client.run("export company")
        client.run("export project")
        client.run("install consumer")
        assert "conanfile.py (consumer/1.0): Msg1:project!!!" in client.out
        assert "conanfile.py (consumer/1.0): Msg2:company!!!" in client.out


def test_multi_top_missing_from_remote():
    """
    https://github.com/conan-io/conan/issues/13656
    """
    tc = TestClient(light=True, default_server_user=True)
    tc.save({"conanfile.py": GenConanfile("base", "1.1")})
    tc.run("create . --user=user --channel=testing")

    tc.save({"conanfile.py": GenConanfile("dep", "1.0")
            .with_python_requires("base/1.1@user/testing")})
    tc.run("create . --user=user --channel=testing -r default")

    tc.run("upload * -c -r default")
    tc.run("remove * -c")

    tc.save({"conanfile.py": GenConanfile("pkg", "1.0")
            .with_python_requires("dep/1.0@user/testing")})

    # This used to crash, with errors about not defining remotes
    tc.run("create . --name=pkg --version=1.0 -r default")

    # Ensure we found them in the remote
    assert "dep/1.0@user/testing: Not found in local cache, looking in remotes..." in tc.out
    assert "dep/1.0@user/testing: Downloaded recipe revision" in tc.out
    assert "base/1.1@user/testing: Not found in local cache, looking in remotes..." in tc.out
    assert "base/1.1@user/testing: Downloaded recipe revision" in tc.out


def test_transitive_range_not_found_in_cache():
    """
    https://github.com/conan-io/conan/issues/13761
    """
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("pr", "1.0")})
    c.run("create .")

    c.save({"conanfile.py": GenConanfile("dep", "1.0").with_python_requires("pr/[>0]")})
    c.run("create .")

    c.save({"conanfile.py": GenConanfile("pkg", "1.0").with_requires("dep/1.0")})
    c.run("create . ")
    c.assert_listed_require({"pr/1.0": "Cache"}, python=True)
    assert "pr/[>0]: pr/1.0" in c.out


def test_export_pkg():
    c = TestClient(light=True)
    c.save({"conanfile.py": GenConanfile("pytool", "0.1").with_package_type("python-require")})
    c.run("export-pkg .", assert_error=True)
    assert "export-pkg can only be used for binaries, not for 'python-require'" in c.out
    # Make sure nothing is exported
    c.run("list *")
    assert "WARN: There are no matching recipe references" in c.out
    assert "pytool/0.1" not in c.out


def test_py_requires_override_method():
    tc = TestClient(light=True)
    pyreq = textwrap.dedent("""
        from conan import ConanFile
        class MyConanfileBase:
            def get_toolchain(self):
                return "mybasetoolchain"

            def generate(self):
                self.output.info("MyConanfileBase generate with value %s" % self.get_toolchain())

        class MyConanfile(ConanFile):
            name = "myconanfile"
            version = "1.0"
            package_type = "python-require"
    """)

    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class MyConanfile(ConanFile):
            name = "app"
            version = "1.0"
            python_requires = "myconanfile/1.0"
            python_requires_extend = "myconanfile.MyConanfileBase"
            def get_toolchain(self):
                return "mycustomtoolchain"
    """)
    tc.save({"myconanfile/conanfile.py": pyreq,
             "conanfile.py": conanfile})
    tc.run("create myconanfile")
    tc.run("create .")
    assert "MyConanfileBase generate with value mycustomtoolchain" in tc.out



import os
import platform
import textwrap
import re

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save, load


@pytest.mark.parametrize("os_", ["Macos", "Linux", "Windows"])
def test_extra_flags_via_conf(os_):
    os_sdk = "tools.apple:sdk_path=/my/sdk/path" if os_ == "Macos" else ""
    profile = textwrap.dedent(f"""
        [settings]
        os={os_}
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release

        [conf]
        tools.build:cxxflags=["--flag1", "--flag2"]
        tools.build:cflags+=["--flag3", "--flag4"]
        tools.build:sharedlinkflags+=["--flag5"]
        tools.build:exelinkflags+=["--flag6"]
        tools.build:defines+=["DEF1", "DEF2"]
        {os_sdk}
        """)
    client = TestClient()
    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type") \
        .with_generator("GnuToolchain")
    client.save({"conanfile.py": conanfile,
                 "profile": profile})
    client.run("install . --profile:build=profile --profile:host=profile")
    toolchain = client.load(
        "conangnutoolchain{}".format('.bat' if os_ == "Windows" else '.sh'))
    if os_ == "Windows":
        assert 'set "CPPFLAGS=%CPPFLAGS% -DNDEBUG -DDEF1 -DDEF2"' in toolchain
        assert 'set "CXXFLAGS=%CXXFLAGS% -O3 --flag1 --flag2"' in toolchain
        assert 'set "CFLAGS=%CFLAGS% -O3 --flag3 --flag4"' in toolchain
        assert 'set "LDFLAGS=%LDFLAGS% --flag5 --flag6"' in toolchain
        assert f'set "PKG_CONFIG_PATH={client.current_folder};%PKG_CONFIG_PATH%"' in toolchain
    elif os_ == "Linux":
        assert 'export CPPFLAGS="$CPPFLAGS -DNDEBUG -DDEF1 -DDEF2"' in toolchain
        assert 'export CXXFLAGS="$CXXFLAGS -O3 --flag1 --flag2"' in toolchain
        assert 'export CFLAGS="$CFLAGS -O3 --flag3 --flag4"' in toolchain
        assert 'export LDFLAGS="$LDFLAGS --flag5 --flag6"' in toolchain
        assert f'export PKG_CONFIG_PATH="{client.current_folder}:$PKG_CONFIG_PATH"' in toolchain
    else:  # macOS
        assert 'export CPPFLAGS="$CPPFLAGS -DNDEBUG -DDEF1 -DDEF2"' in toolchain
        assert 'export CXXFLAGS="$CXXFLAGS -O3 --flag1 --flag2"' in toolchain
        assert 'export CFLAGS="$CFLAGS -O3 --flag3 --flag4"' in toolchain
        assert 'export LDFLAGS="$LDFLAGS --flag5 --flag6"' in toolchain
        assert f'export PKG_CONFIG_PATH="{client.current_folder}:$PKG_CONFIG_PATH"' in toolchain


def test_extra_flags_order():
    client = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.gnu import GnuToolchain

        class Conan(ConanFile):
            name = "pkg"
            version = "0.1"
            settings = "os", "arch", "build_type"
            def generate(self):
                at = GnuToolchain(self)
                at.extra_cxxflags = ["extra_cxxflags"]
                at.extra_cflags = ["extra_cflags"]
                at.extra_ldflags = ["extra_ldflags"]
                at.extra_defines = ["extra_defines"]
                at.generate()
        """)
    profile = textwrap.dedent("""
        include(default)
        [conf]
        tools.build:cxxflags+=['cxxflags']
        tools.build:cflags+=['cflags']
        tools.build:sharedlinkflags+=['sharedlinkflags']
        tools.build:exelinkflags+=['exelinkflags']
        tools.build:defines+=['defines']
        """)
    client.save({"conanfile.py": conanfile, "profile": profile})
    client.run('install . -pr=./profile')
    toolchain = client.load(
        "conangnutoolchain{}".format('.bat' if platform.system() == "Windows" else '.sh'))

    assert '-Dextra_defines -Ddefines' in toolchain
    assert 'extra_cxxflags cxxflags' in toolchain
    assert 'extra_cflags cflags' in toolchain
    assert 'extra_ldflags sharedlinkflags exelinkflags' in toolchain


def test_autotools_custom_environment():
    client = TestClient()
    conanfile = textwrap.dedent("""
            from conan import ConanFile
            from conan.tools.gnu import GnuToolchain

            class Conan(ConanFile):
                settings = "os"
                def generate(self):
                    at = GnuToolchain(self)
                    env = at.extra_env
                    env.define("FOO", "BAR")
                    at.generate()
            """)

    client.save({"conanfile.py": conanfile})
    client.run("install . -s:b os=Linux -s:h os=Linux")
    content = load(os.path.join(client.current_folder, "conangnutoolchain.sh"))
    assert 'export FOO="BAR"' in content


@pytest.mark.parametrize("os_", ["Linux", "Windows"])
def test_linker_scripts_via_conf(os_):
    profile = textwrap.dedent("""
        [settings]
        os=%s
        compiler=gcc
        compiler.version=6
        compiler.libcxx=libstdc++11
        arch=armv8
        build_type=Release

        [conf]

        tools.build:sharedlinkflags+=["--flag5"]
        tools.build:exelinkflags+=["--flag6"]
        tools.build:linker_scripts+=["/linker/scripts/flash.ld", "/linker/scripts/extra_data.ld"]
        """ % os_)
    client = TestClient()
    conanfile = GenConanfile().with_settings("os", "arch", "compiler", "build_type") \
        .with_generator("GnuToolchain")
    client.save({"conanfile.py": conanfile,
                 "profile": profile})
    client.run("install . --profile:build=profile --profile:host=profile")
    toolchain = client.load(
        "conangnutoolchain{}".format('.bat' if os_ == "Windows" else '.sh'))
    if os_ == "Windows":
        assert 'set "LDFLAGS=%LDFLAGS% --flag5 --flag6 -T\'/linker/scripts/flash.ld\' -T\'/linker/scripts/extra_data.ld\'"' in toolchain
    else:
        assert 'export LDFLAGS="$LDFLAGS --flag5 --flag6 -T\'/linker/scripts/flash.ld\' -T\'/linker/scripts/extra_data.ld\'"' in toolchain


def test_not_none_values():
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.gnu import GnuToolchain

        class Foo(ConanFile):
            name = "foo"
            version = "1.0"

            def generate(self):
                tc = GnuToolchain(self)
                assert None not in tc.defines
                assert None not in tc.cxxflags
                assert None not in tc.cflags
                assert None not in tc.ldflags

    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("install .")


def test_set_prefix():
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.gnu import GnuToolchain
        from conan.tools.layout import basic_layout


        class Foo(ConanFile):
            name = "foo"
            version = "1.0"
            def layout(self):
                basic_layout(self)
            def generate(self):
                at_toolchain = GnuToolchain(self, prefix="/somefolder")
                at_toolchain.generate()
    """)

    client = TestClient()
    client.save({"conanfile.py": conanfile})
    client.run("install .")
    conanbuild = client.load(
        os.path.join(client.current_folder, "build", "conan", "conanbuild.conf"))
    assert "--prefix=/somefolder" in conanbuild
    assert conanbuild.count("--prefix") == 1


def test_unknown_compiler():
    client = TestClient()
    save(client.paths.settings_path_user, "compiler:\n  xlc:\n")
    client.save({"conanfile.py": GenConanfile().with_settings("compiler", "build_type")
                .with_generator("GnuToolchain")
                 })
    # this used to crash, because of build_type_flags in GnuToolchain returning empty string
    client.run("install . -s compiler=xlc")
    assert "conanfile.py: Generator 'GnuToolchain' calling 'generate()'" in client.out


def test_toolchain_and_compilers_build_context():
    """
    Tests how GnuToolchain manages the build context profile if the build profile is
    specifying another compiler path (using conf)

    Issue related: https://github.com/conan-io/conan/issues/15878
    """
    host = textwrap.dedent("""
    [settings]
    arch=armv8
    build_type=Release
    compiler=gcc
    compiler.cppstd=gnu17
    compiler.libcxx=libstdc++11
    compiler.version=11
    os=Linux

    [conf]
    tools.build:compiler_executables={"c": "gcc", "cpp": "g++", "rc": "windres"}
    """)
    build = textwrap.dedent("""
    [settings]
    os=Linux
    arch=x86_64
    compiler=clang
    compiler.version=12
    compiler.libcxx=libc++
    compiler.cppstd=11

    [conf]
    tools.build:compiler_executables={"c": "clang", "cpp": "clang++"}
    """)
    tool = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class toolRecipe(ConanFile):
        name = "tool"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"
        generators = "GnuToolchain"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conangnutoolchain.sh")
            content = load(self, toolchain)
            assert 'export CC="clang"' in content
            assert 'export CXX="clang++"' in content
    """)
    consumer = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class consumerRecipe(ConanFile):
        name = "consumer"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"
        generators = "GnuToolchain"
        tool_requires = "tool/1.0"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conangnutoolchain.sh")
            content = load(self, toolchain)
            assert 'export CC="gcc"' in content
            assert 'export CXX="g++"' in content
            assert 'export RC="windres"' in content
            # Issue: https://github.com/conan-io/conan/issues/15486
            assert 'export CC_FOR_BUILD="clang"' in content
            assert 'export CXX_FOR_BUILD="clang++"' in content
    """)
    client = TestClient()
    client.save({
        "host": host,
        "build": build,
        "tool/conanfile.py": tool,
        "consumer/conanfile.py": consumer
    })
    client.run("export tool")
    client.run("create consumer -pr:h host -pr:b build --build=missing")


def test_autotools_crossbuild_ux():
    client = TestClient()
    profile_build = textwrap.dedent("""
        [settings]
        os = Macos
        os.version=10.11
        arch = armv7
        compiler = apple-clang
        compiler.version = 12.0
        compiler.libcxx = libc++
        """)
    profile_host = textwrap.dedent("""
        [settings]
        arch=x86_64
        build_type=Release
        compiler=apple-clang
        compiler.cppstd=gnu17
        compiler.libcxx=libc++
        compiler.version=15
        os=Macos
        [conf]
        tools.apple:sdk_path=/my/sdk/path
        """)
    conanfile = textwrap.dedent("""
    from conan import ConanFile
    from conan.tools.gnu import GnuToolchain
    from conan.tools.build import cross_building
    class Pkg(ConanFile):
        settings = "os", "arch", "compiler", "build_type"
        def generate(self):
            tc = GnuToolchain(self)
            if cross_building(self):
                host_arch = tc.triplets_info["host"]["machine"]
                build_arch = tc.triplets_info["build"]["machine"]
                if host_arch and build_arch:
                    tc.configure_args["--host"] = f"{host_arch}-my-triplet-host"
                    tc.configure_args["--build"] = f"{build_arch}-my-triplet-build"
            tc.generate()
            """)

    client.save({"conanfile.py": conanfile,
                 "profile_build": profile_build,
                 "profile_host": profile_host})
    client.run("install . --profile:build=profile_build --profile:host=profile_host")
    conanbuild = client.load("conanbuild.conf")
    host_flags = re.findall(r"--host=[\w-]*\b", conanbuild)
    build_flags = re.findall(r"--build=[\w-]*\b", conanbuild)
    assert len(host_flags) == 1
    assert len(build_flags) == 1
    assert host_flags[0] == '--host=x86_64-my-triplet-host'
    assert build_flags[0] == '--build=arm-my-triplet-build'


def test_msvc_profile_defaults():
    """
    Tests how GnuToolchain manages MSVC profile and its default env variables.
    """
    profile = textwrap.dedent("""\
    [settings]
    os=Windows
    arch=x86_64
    compiler=msvc
    compiler.version=191
    compiler.runtime=dynamic
    build_type=Release

    [conf]
    tools.build:compiler_executables={"c": "clang", "cpp": "clang++"}
    # Fake installation path
    tools.microsoft.msbuild:installation_path={{os.getcwd()}}
    """)
    # Consumer with default values
    consumer = textwrap.dedent("""\
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class consumerRecipe(ConanFile):
        name = "consumer"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"
        generators = "GnuToolchain"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conangnutoolchain.bat")
            content = load(self, toolchain)
            # Default values and conf ones
            assert r'set "CC=clang"' in content  # conf value has precedence
            assert r'set "CXX=clang++"' in content  # conf value has precedence
            assert 'set "NM=dumpbin -symbols"' in content
            assert 'set "OBJDUMP=:"' in content
            assert 'set "RANLIB=:"' in content
            assert 'set "STRIP=:"' in content
    """)
    client = TestClient()
    client.save({
        "profile": profile,
        "consumer/conanfile.py": consumer
    })
    client.run("create consumer -pr:a profile --build=missing")
    # Consumer changing default values
    consumer = textwrap.dedent("""\
    import os
    from conan import ConanFile
    from conan.tools.files import load
    from conan.tools.gnu import GnuToolchain

    class consumerRecipe(ConanFile):
        name = "consumer"
        version = "1.0"
        # Binary configuration
        settings = "os", "compiler", "build_type", "arch"

        def generate(self):
            tc = GnuToolchain(self)
            # Prepending compiler wrappers
            tc.extra_env.prepend("CC", "compile")
            tc.extra_env.prepend("CXX", "compile")
            tc.extra_env.define("OBJDUMP", "other-value")
            tc.extra_env.unset("RANLIB")
            tc.generate()

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conangnutoolchain.bat")
            content = load(self, toolchain)
            # Default values
            assert r'set "CC=compile clang"' in content
            assert r'set "CXX=compile clang++"' in content
            assert 'set "NM=dumpbin -symbols"' in content
            assert 'set "OBJDUMP=other-value"' in content  # redefined
            assert 'set "RANLIB=:"' not in content  # removed
            assert 'set "STRIP=:"' in content
    """)
    client.save({
        "consumer/conanfile.py": consumer
    })
    client.run("create consumer -pr:a profile --build=missing")


def test_conf_build_does_not_exist():
    host = textwrap.dedent("""
    [settings]
    arch=x86_64
    os=Linux
    [conf]
    tools.build:compiler_executables={'c': '/usr/bin/gcc', 'cpp': '/usr/bin/g++'}
    """)
    build = textwrap.dedent("""
    [settings]
    arch=armv8
    os=Linux
    [conf]
    tools.build:compiler_executables={'c': 'x86_64-linux-gnu-gcc', 'cpp': 'x86_64-linux-gnu-g++'}
    """)
    c = TestClient()
    c.save({"conanfile.py": GenConanfile("pkg", "0.1"),
            "host": host,
            "build": build})
    c.run("export .")
    c.run("install --requires=pkg/0.1 --build=pkg/0.1 -g GnuToolchain -pr:h host -pr:b build")
    tc = c.load("conangnutoolchain.sh")
    assert 'export CC_FOR_BUILD="x86_64-linux-gnu-gcc"' in tc
    assert 'export CXX_FOR_BUILD="x86_64-linux-gnu-g++"' in tc


@pytest.mark.parametrize("toolchain", ["GnuToolchain", "AutotoolsToolchain"])
def test_conf_extra_apple_flags(toolchain):
    host = textwrap.dedent("""
    [settings]
    arch=x86_64
    os=Macos
    [conf]
    tools.apple:enable_bitcode = True
    tools.apple:enable_arc = True
    tools.apple:enable_visibility = True
    """)

    c = TestClient()
    c.save({"conanfile.txt": f"[generators]\n{toolchain}",
            "host": host})
    c.run("install . -pr:a host")
    f = "conanautotoolstoolchain.sh" if toolchain == "AutotoolsToolchain" else "conangnutoolchain.sh"
    tc = c.load(f)
    assert 'export CXXFLAGS="$CXXFLAGS -fembed-bitcode -fvisibility=default"' in tc
    assert 'export CFLAGS="$CFLAGS -fembed-bitcode -fvisibility=default"' in tc
    assert 'export LDFLAGS="$LDFLAGS -fembed-bitcode -fvisibility=default"' in tc
    assert 'export OBJCFLAGS="$OBJCFLAGS -fobjc-arc"' in tc
    assert 'export OBJCXXFLAGS="$OBJCXXFLAGS -fobjc-arc"' in tc

    c.run("install . -pr:a host -s build_type=Debug")
    tc = c.load(f)
    assert 'export CXXFLAGS="$CXXFLAGS -fembed-bitcode-marker -fvisibility=default"' in tc
    assert 'export CFLAGS="$CFLAGS -fembed-bitcode-marker -fvisibility=default"' in tc
    assert 'export LDFLAGS="$LDFLAGS -fembed-bitcode-marker -fvisibility=default"' in tc
    assert 'export OBJCFLAGS="$OBJCFLAGS -fobjc-arc"' in tc
    assert 'export OBJCXXFLAGS="$OBJCXXFLAGS -fobjc-arc"' in tc

    host = textwrap.dedent("""
        [settings]
        arch=x86_64
        os=Macos
        [conf]
        tools.apple:enable_bitcode = False
        tools.apple:enable_arc = False
        tools.apple:enable_visibility = False
        """)

    c.save({"host": host})
    c.run("install . -pr:a host")
    tc = c.load(f)
    assert 'CXXFLAGS="$CXXFLAGS -fvisibility=hidden -fvisibility-inlines-hidden"' in tc
    assert 'CFLAGS="$CFLAGS -fvisibility=hidden -fvisibility-inlines-hidden"' in tc
    assert 'LDFLAGS="$LDFLAGS -fvisibility=hidden -fvisibility-inlines-hidden"' in tc
    assert 'export OBJCFLAGS="$OBJCFLAGS -fno-objc-arc"' in tc
    assert 'export OBJCXXFLAGS="$OBJCXXFLAGS -fno-objc-arc"' in tc


def test_toolchain_crossbuild_to_android():
    """
    Issue related: https://github.com/conan-io/conan/issues/17441
    """
    build = textwrap.dedent("""
    [settings]
    arch=armv8
    build_type=Release
    compiler=gcc
    compiler.cppstd=gnu17
    compiler.libcxx=libstdc++11
    compiler.version=11
    os=Linux
    """)
    host = textwrap.dedent("""
    [settings]
    os = Android
    os.api_level = 21
    arch=x86_64
    compiler=clang
    compiler.version=12
    compiler.libcxx=libc++
    compiler.cppstd=11

    [buildenv]
    CC=clang
    CXX=clang++

    [conf]
    tools.android:ndk_path=/path/to/ndk
    """)
    consumer = textwrap.dedent("""
    import os
    from conan import ConanFile
    from conan.tools.files import load

    class consumerRecipe(ConanFile):
        name = "consumer"
        version = "1.0"
        settings = "os", "compiler", "build_type", "arch"
        generators = "GnuToolchain"

        def build(self):
            toolchain = os.path.join(self.generators_folder, "conangnutoolchain.sh")
            content = load(self, toolchain)
            assert 'export CC="clang"' not in content
            assert 'export CXX="clang++"' not in content
            assert 'export LD="/path/to/ndk' in content
            assert 'export STRIP="/path/to/ndk' in content
            assert 'export RANLIB="/path/to/ndk' in content
            assert 'export AS="/path/to/ndk' in content
            assert 'export AR="/path/to/ndk' in content
            assert 'export ADDR2LINE="/path/to/ndk' in content
            assert 'export NM="/path/to/ndk' in content
            assert 'export OBJCOPY="/path/to/ndk' in content
            assert 'export OBJDUMP="/path/to/ndk' in content
            assert 'export READELF="/path/to/ndk' in content
            assert 'export ELFEDIT="/path/to/ndk' in content

            build_env = os.path.join(self.generators_folder, "conanbuildenv-x86_64.sh")
            content = load(self, build_env)
            assert 'export CC="clang"' in content
            assert 'export CXX="clang++"' in content
            assert 'export LD=' not in content
            assert 'export STRIP="/path/to/ndk' not in content
            assert 'export RANLIB="/path/to/ndk' not in content
            assert 'export AS="/path/to/ndk' not in content
            assert 'export AR="/path/to/ndk' not in content
            assert 'export ADDR2LINE="/path/to/ndk' not in content
            assert 'export NM="/path/to/ndk' not in content
            assert 'export OBJCOPY="/path/to/ndk' not in content
            assert 'export OBJDUMP="/path/to/ndk' not in content
            assert 'export READELF="/path/to/ndk' not in content
            assert 'export ELFEDIT="/path/to/ndk' not in content
    """)
    client = TestClient()
    client.save({
        "host": host,
        "build": build,
        "conanfile.py": consumer
    })
    client.run("create . -pr:h host -pr:b build")


@pytest.mark.parametrize(
    "threads, flags",
    [("posix", "-pthread"), ("wasm_workers", "-sWASM_WORKERS=1")],
)
def test_thread_flags(threads, flags):
    client = TestClient()
    profile = textwrap.dedent(f"""
        [settings]
        arch=wasm
        build_type=Release
        compiler=emcc
        compiler.cppstd=17
        compiler.threads={threads}
        compiler.libcxx=libc++
        compiler.version=4.0.10
        os=Emscripten
        """)
    client.save(
        {
            "conanfile.py": GenConanfile("pkg", "1.0")
            .with_settings("os", "arch", "compiler", "build_type")
            .with_generator("GnuToolchain"),
            "profile": profile,
        }
    )
    client.run("install . -pr=./profile")
    os = platform.system()
    toolchain = client.load(
        "conangnutoolchain{}".format('.bat' if os == "Windows" else '.sh'))

    if os == "Windows":
        assert f'set "CXXFLAGS=%CXXFLAGS% -stdlib=libc++ {flags}"' in toolchain
        assert f'set "CFLAGS=%CFLAGS% {flags}"' in toolchain
        assert f'set "LDFLAGS=%LDFLAGS% {flags}' in toolchain
    else:
        assert f'export CXXFLAGS="$CXXFLAGS -stdlib=libc++ {flags}"' in toolchain
        assert f'export CFLAGS="$CFLAGS {flags}"' in toolchain
        assert f'export LDFLAGS="$LDFLAGS {flags}"' in toolchain



import calendar
import datetime
import time

from dateutil import parser


def from_timestamp_to_iso8601(timestamp):
    # Used exclusively by conan_server to return the date in iso format (same as artifactory)
    return "%s" % datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc).isoformat()


def _from_iso8601_to_datetime(iso_str):
    return parser.isoparse(iso_str)


def from_iso8601_to_timestamp(iso_str):
    # used by RestClient v2 to transform from HTTP API (iso) to Conan internal timestamp
    datetime_time = _from_iso8601_to_datetime(iso_str)
    return datetime_time.timestamp()


def timestamp_now():
    # seconds since epoch 0, easy to store, in UTC
    # Used in Manifest timestamp, in packagesDB LRU and in timestamp of backup-sources json
    return calendar.timegm(time.gmtime())


def revision_timestamp_now():
    return time.time()


def timestamp_to_str(timestamp):
    # used by ref.repr_humantime() to print human readable time
    assert timestamp is not None
    return datetime.datetime.fromtimestamp(int(timestamp), datetime.timezone.utc).strftime('%Y-%m-%d %H:%M:%S UTC')



from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_version_range_conf_nonexplicit_expression():
    tc = TestClient(light=True)

    tc.save({"base/conanfile.py": GenConanfile("base")})
    tc.run("create base/conanfile.py --version=1.5.1")
    tc.run("create base/conanfile.py --version=2.5.0-pre")

    tc.save({"v1/conanfile.py": GenConanfile("pkg", "1.0").with_requires("base/[>1 <2]"),
             "v2/conanfile.py": GenConanfile("pkg", "2.0").with_requires("base/[>2 <3]")})

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=False"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2]: base/1.5.1" in tc.out
    tc.run("create v2/conanfile.py", assert_error=True)
    assert "Package 'base/[>2 <3]' not resolved" in tc.out

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=True"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2]: base/1.5.1" in tc.out
    tc.run("create v2/conanfile.py")
    assert "base/[>2 <3]: base/2.5.0-pre" in tc.out

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=None"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2]: base/1.5.1" in tc.out

    tc.run("create v2/conanfile.py", assert_error=True)
    assert "Package 'base/[>2 <3]' not resolved" in tc.out


def test_version_range_conf_explicit_expression():
    tc = TestClient(light=True)

    tc.save({"base/conanfile.py": GenConanfile("base")})
    tc.run("create base/conanfile.py --version=1.5.1")
    tc.run("create base/conanfile.py --version=2.5.0-pre")

    tc.save({"v1/conanfile.py": GenConanfile("pkg", "1.0").with_requires("base/[>1 <2, include_prerelease]"),
             "v2/conanfile.py": GenConanfile("pkg", "2.0").with_requires("base/[>2 <3, include_prerelease]")})

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=False"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2, include_prerelease]: base/1.5.1" in tc.out
    tc.run("create v2/conanfile.py", assert_error=True)
    assert "Package 'base/[>2 <3, include_prerelease]' not resolved" in tc.out

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=True"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2, include_prerelease]: base/1.5.1" in tc.out
    tc.run("create v2/conanfile.py")
    assert "base/[>2 <3, include_prerelease]: base/2.5.0-pre" in tc.out

    tc.save_home({"global.conf": "core.version_ranges:resolve_prereleases=None"})
    tc.run("create v1/conanfile.py")
    assert "base/[>1 <2, include_prerelease]: base/1.5.1" in tc.out
    tc.run("create v2/conanfile.py")
    assert "base/[>2 <3, include_prerelease]: base/2.5.0-pre" in tc.out



import textwrap

from conan.test.utils.tools import TestClient


class TestExportFoldersAvailability:

    def test_export_sources_folder_availability_local_methods(self):
        conanfile = textwrap.dedent('''
        import os
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                assert os.path.exists(self.export_sources_folder)

            def export(self):
                assert self.export_sources_folder is None

            def export_sources(self):
                assert os.path.exists(self.export_sources_folder)

            def source(self):
                assert os.path.exists(self.export_sources_folder)

            def build(self):
                assert os.path.exists(self.export_sources_folder)

            def package(self):
                assert os.path.exists(self.export_sources_folder)

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})

        client.run("export . --name foo --version 1.0")
        client.run("install .")
        client.run("source .")
        client.run("build .")

    def test_export_folder_availability_local_methods(self):
        conanfile = textwrap.dedent('''
        import os
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                assert os.path.exists(self.export_sources_folder)

            def export_sources(self):
                # We need it available for the post_export hook so it is available
                assert os.path.exists(self.export_folder)

            def export(self):
                assert os.path.exists(self.export_folder)

            def source(self):
                assert os.path.exists(self.export_sources_folder)

            def build(self):
                assert os.path.exists(self.export_sources_folder)

            def package(self):
                assert os.path.exists(self.export_sources_folder)

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})

        client.run("export . --name foo --version 1.0")
        client.run("install .")
        client.run("source .")
        client.run("build .")

    def test_export_folder_availability_create(self):
        conanfile = textwrap.dedent('''
        import os
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                assert self.export_folder is None

            def export(self):
                assert os.path.exists(self.export_folder)

            def export_sources(self):
                # We need it available for the post_export hook so it is available
                assert os.path.exists(self.export_folder)

            def source(self):
                assert self.export_folder is None

            def build(self):
                assert self.export_folder is None

            def package(self):
                assert self.export_folder is None

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . --name foo --version 1.0")

    def test_export_sources_folder_availability_create(self):
        conanfile = textwrap.dedent('''
        import os
        from conan import ConanFile

        class ConanLib(ConanFile):

            def layout(self):
                self.folders.source = "MY_SOURCE"

            def generate(self):
                assert os.path.exists(self.export_sources_folder)

            def export(self):
                assert self.export_sources_folder is None

            def export_sources(self):
                assert os.path.exists(self.export_sources_folder)

            def source(self):
                assert os.path.exists(self.export_sources_folder)

            def build(self):
                assert os.path.exists(self.export_sources_folder)

            def package(self):
                assert os.path.exists(self.export_sources_folder)

        ''')
        client = TestClient()
        client.save({"conanfile.py": conanfile})
        client.run("create . --name foo --version 1.0")



import os
import textwrap

import pytest

from conan import __version__
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient
from conan.internal.util.files import save


class TestNewCommand:
    def test_new_cmake_lib(self):
        client = TestClient(light=True)
        client.run("new cmake_lib -d name=pkg -d version=1.3")
        conanfile = client.load("conanfile.py")
        assert "CMakeToolchain" in conanfile
        conanfile = client.load("test_package/conanfile.py")
        assert "CMakeToolchain" in conanfile
        cmake = client.load("test_package/CMakeLists.txt")
        assert "find_package" in cmake

        # Test at least it exports correctly
        client.run("export . --user=myuser --channel=testing")
        assert "pkg/1.3@myuser/testing" in client.out

    def test_new_cmake_exe(self):
        client = TestClient(light=True)
        client.run("new cmake_exe -d name=pkg -d version=1.3")
        conanfile = client.load("conanfile.py")
        assert "CMakeToolchain" in conanfile
        conanfile = client.load("test_package/conanfile.py")
        assert "def test(self):" in conanfile

        # Test at least it exports correctly
        client.run("export . --user=myuser --channel=testing")
        assert "pkg/1.3@myuser/testing" in client.out

    def test_new_basic_template(self):
        tc = TestClient(light=True)
        tc.run("new basic")
        assert '# self.requires("zlib/1.2.13")' in tc.load("conanfile.py")

        tc.run("new basic -d name=mygame -d requires=math/1.0 -d requires=ai/1.0 -f")
        conanfile = tc.load("conanfile.py")
        assert 'self.requires("math/1.0")' in conanfile
        assert 'self.requires("ai/1.0")' in conanfile
        assert 'name = "mygame"' in conanfile

    def test_new_defaults(self):
        c = TestClient(light=True)
        for t in ("cmake_lib", "cmake_exe", "meson_lib", "meson_exe", "msbuild_lib", "msbuild_exe",
                  "bazel_lib", "bazel_exe", "autotools_lib", "autotools_exe"):
            c.save({}, clean_first=True)
            c.run(f"new {t}")
            conanfile = c.load("conanfile.py")
            assert 'name = "mypkg"' in conanfile
            assert 'version = "0.1"' in conanfile
            if t == "cmake_lib":
                cmake = c.load("CMakeLists.txt")
                assert "src/mypkg.cpp" in cmake
                cpp = c.load("src/mypkg.cpp")
                assert "mypkg()" in cpp
                hpp = c.load("include/mypkg.h")
                assert "void mypkg();" in hpp

            c.run(f"new {t} -d name=mylib -f")
            conanfile = c.load("conanfile.py")
            assert 'name = "mylib"' in conanfile
            assert 'version = "0.1"' in conanfile


class TestNewCommandUserTemplate:

    @pytest.mark.parametrize("folder", ("mytemplate", "sub/mytemplate"))
    def test_user_template(self, folder):
        client = TestClient(light=True)
        template1 = textwrap.dedent("""
            class Conan(ConanFile):
                name = "{{name}}"
                version = "{{version}}"
                conan_version = "{{conan_version}}"
        """)
        save(os.path.join(client.cache_folder, f"templates/command/new/{folder}/conanfile.py"),
             template1)
        client.run(f"new {folder} -d name=hello -d version=0.1")
        conanfile = client.load("conanfile.py")
        assert 'name = "hello"' in conanfile
        assert 'version = "0.1"' in conanfile
        assert 'conan_version = "{}"'.format(__version__) in conanfile

    def test_user_template_abs(self):
        tmp_folder = temp_folder()
        client = TestClient(light=True)
        template1 = textwrap.dedent("""
            class Conan(ConanFile):
                name = "{{name}}"
        """)
        save(os.path.join(tmp_folder, "conanfile.py"), template1)
        client.run(f'new "{tmp_folder}" -d name=hello -d version=0.1')
        conanfile = client.load("conanfile.py")
        assert 'name = "hello"' in conanfile

    def test_user_template_filenames(self):
        client = TestClient(light=True)
        save(os.path.join(client.cache_folder, "templates/command/new/mytemplate/{{name}}"), "Hi!")
        client.run(f"new mytemplate -d name=pkg.txt")
        assert "Hi!" == client.load("pkg.txt")

    def test_skip_files(self):
        client = TestClient(light=True)
        template1 = textwrap.dedent("""
            class Conan(ConanFile):
                name = "{{name}}"
                version = "{{version}}"
                conan_version = "{{conan_version}}"
        """)
        path = os.path.join(client.cache_folder, f"templates/command/new/mytemplate")
        save(os.path.join(path, "conanfile.py"), template1)
        save(os.path.join(path, "file.h"), "{{header}}")

        client.run(f"new mytemplate -d name=hello -d version=0.1 -d header=")
        assert not os.path.exists(os.path.join(client.current_folder, "file.h"))
        assert not os.path.exists(os.path.join(client.current_folder, "file.cpp"))
        client.run(f"new mytemplate -d name=hello -d version=0.1 -d header=xxx -f")
        assert os.path.exists(os.path.join(client.current_folder, "file.h"))
        assert not os.path.exists(os.path.join(client.current_folder, "file.cpp"))

    def test_template_image_files(self):
        """ problematic files that we dont want to render with Jinja, like PNG or other binaries,
        have to be explicitly excluded from render"""
        client = TestClient(light=True)
        template_dir = "templates/command/new/t_dir"
        png = "$&(){}{}{{}{}"
        save(os.path.join(client.cache_folder, template_dir, "myimage.png"), png)
        client.run("new t_dir -d name=hello -d version=0.1", assert_error=True)
        assert "TemplateSyntaxError" in client.out

        save(os.path.join(client.cache_folder, template_dir, "not_templates"), "*.png")
        client.run("new t_dir -d name=hello -d version=0.1")
        myimage = client.load("myimage.png")
        assert myimage == png
        assert not os.path.exists(os.path.join(client.current_folder, "not_templates"))


class TestNewErrors:
    def test_template_errors(self):
        client = TestClient(light=True)
        client.run("new mytemplate", assert_error=True)
        assert "ERROR: Template doesn't exist" in client.out

    def test_forced(self):
        client = TestClient(light=True)
        client.run("new cmake_lib -d name=hello -d version=0.1")
        client.run("new cmake_lib -d name=hello -d version=0.1", assert_error=True)
        client.run("new cmake_lib -d name=bye -d version=0.2 --force")
        conanfile = client.load("conanfile.py")
        assert 'name = "bye"' in conanfile
        assert 'version = "0.2"' in conanfile

    def test_duplicated(self):
        client = TestClient(light=True)
        client.run("new cmake_lib -d name=hello -d name=0.1", assert_error=True)
        assert "ERROR: name argument can't be multiple: ['hello', '0.1']" in client.out

        # It will create a list and assign to it, but it will not fail ugly
        client.run("new cmake_lib -d name=pkg -d version=0.1 -d version=0.2", assert_error=True)
        assert "ERROR: version argument can't be multiple: ['0.1', '0.2']" in client.out

    def test_name_uppercase(self):
        client = TestClient(light=True)
        client.run("new cmake_lib -d name=Hello", assert_error=True)
        assert "ERROR: name argument must be lowercase: Hello" in client.out

    def test_new_change_folder(self):
        client = TestClient(light=True)
        client.run("new cmake_lib -d name=hello -d version=0.1 -o=myfolder")
        assert os.path.exists(os.path.join(client.current_folder, "myfolder", "conanfile.py"))



import os
import platform
import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.assets.sources import gen_function_cpp
from conan.test.utils.tools import TestClient

from conan.internal.util.files import save


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
def test_custom_config_generator_multi():
    conanfile = textwrap.dedent("""
        import os
        from conan import ConanFile
        from conan.tools.cmake import CMakeDeps
        from conan.tools.files import copy
        class App(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            requires = "hello/0.1"

            def generate(self):
                cmake = CMakeDeps(self)
                if self.dependencies["hello"].options.shared:
                    cmake.configuration = "ReleaseShared"
                cmake.generate()

                config = str(self.settings.build_type)
                if self.dependencies["hello"].options.shared:
                    config = "ReleaseShared"
                src = os.path.join(self.dependencies["hello"].package_folder, "bin")
                dst = os.path.join(self.build_folder, config)
                copy(self, "*.dll", src, dst, keep_path=False)
        """)

    app = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

    cmakelist = textwrap.dedent("""
        set(CMAKE_CONFIGURATION_TYPES Debug Release ReleaseShared CACHE STRING
            "Available build-types: Debug, Release and ReleaseShared")

        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 2.8)
        project(App C CXX)

        set(CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_PREFIX_PATH})
        set(CMAKE_MODULE_PATH ${CMAKE_BINARY_DIR} ${CMAKE_MODULE_PATH})

        set(CMAKE_CXX_FLAGS_RELEASESHARED ${CMAKE_CXX_FLAGS_RELEASE})
        set(CMAKE_C_FLAGS_RELEASESHARED ${CMAKE_C_FLAGS_RELEASE})
        set(CMAKE_EXE_LINKER_FLAGS_RELEASESHARED ${CMAKE_EXE_LINKER_FLAGS_RELEASE})

        find_package(hello REQUIRED)
        add_executable(app app.cpp)
        target_link_libraries(app PRIVATE hello::hello)
        """)

    c = TestClient(path_with_spaces=False)
    c.run("new cmake_lib -d name=hello -d version=0.1")
    c.run("create . -s compiler.version=191 "
          "-s build_type=Release -o hello/*:shared=True -tf=\"\"")
    c.run("create . --name=hello --version=0.1 -s compiler.version=191 "
          "-s build_type=Release -tf=\"\"")

    # Prepare the actual consumer package
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmakelist,
            "app.cpp": app})

    settings = {"compiler": "msvc",
                "compiler.version": "191",
                "compiler.runtime": "dynamic",
                "arch": "x86_64",
                "build_type": "Release",
                }

    settings = " ".join('-s %s="%s"' % (k, v) for k, v in settings.items() if v)

    # Run the configure corresponding to this test case
    with c.chdir('build'):
        c.run("install .. %s -o hello/*:shared=True -of=." % settings)
        c.run("install .. %s -o hello/*:shared=False -of=." % settings)
        assert os.path.isfile(os.path.join(c.current_folder,
                                           "hello-Target-releaseshared.cmake"))
        assert os.path.isfile(os.path.join(c.current_folder,
                                           "hello-Target-release.cmake"))

        c.run_command('cmake .. -G "Visual Studio 15 Win64" --loglevel=DEBUG')
        assert "Found DLL and STATIC" in c.out
        c.run_command('cmake --build . --config Release')
        c.run_command(r"Release\\app.exe")
        assert "hello/0.1: Hello World Release!" in c.out
        assert "main: Release!" in c.out
        c.run_command('cmake --build . --config ReleaseShared')
        c.run_command(r"ReleaseShared\\app.exe")
        assert "hello/0.1: Hello World Release!" in c.out
        assert "main: Release!" in c.out


@pytest.mark.tool("cmake")
@pytest.mark.skipif(platform.system() != "Windows", reason="Only for windows")
def test_custom_config_settings():
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        from conan.tools.cmake import CMakeDeps

        class App(ConanFile):
            settings = "os", "arch", "compiler", "build_type"
            requires = "hello/0.1"
            generators = "CMakeToolchain"

            def generate(self):
                cmake = CMakeDeps(self)
                #cmake.configurations.append("MyRelease") # NOT NECESSARY!!!
                cmake.generate()
        """)

    app = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])

    cmakelist = textwrap.dedent("""
        set(CMAKE_CONFIGURATION_TYPES Debug Release MyRelease CACHE STRING
            "Available build-types: Debug, Release and MyRelease")

        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        set(CMAKE_C_COMPILER_WORKS 1)
        set(CMAKE_C_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(App C CXX)

        set(CMAKE_PREFIX_PATH ${CMAKE_BINARY_DIR} ${CMAKE_PREFIX_PATH})
        set(CMAKE_MODULE_PATH ${CMAKE_BINARY_DIR} ${CMAKE_MODULE_PATH})

        set(CMAKE_CXX_FLAGS_MYRELEASE ${CMAKE_CXX_FLAGS_RELEASE})
        set(CMAKE_C_FLAGS_MYRELEASE ${CMAKE_C_FLAGS_RELEASE})
        set(CMAKE_EXE_LINKER_FLAGS_MYRELEASE ${CMAKE_EXE_LINKER_FLAGS_RELEASE})

        find_package(hello REQUIRED)
        add_executable(app app.cpp)
        target_link_libraries(app PRIVATE hello::hello)
        """)

    c = TestClient(path_with_spaces=False)
    settings = "build_type: [MyRelease]"
    save(c.paths.settings_path_user, settings)
    c.run("new cmake_lib -d name=hello -d version=0.1")
    cmake = c.load("CMakeLists.txt")

    cmake = cmake.replace("cmake_minimum_required", """
        set(CMAKE_CONFIGURATION_TYPES Debug MyRelease Release CACHE STRING "Types")

        cmake_minimum_required""")
    cmake = cmake.replace("add_library", textwrap.dedent("""
        set(CMAKE_CXX_FLAGS_MYRELEASE ${CMAKE_CXX_FLAGS_RELEASE})
        set(CMAKE_C_FLAGS_MYRELEASE ${CMAKE_C_FLAGS_RELEASE})
        set(CMAKE_EXE_LINKER_FLAGS_MYRELEASE ${CMAKE_EXE_LINKER_FLAGS_RELEASE})
        add_library"""))
    cmake = cmake.replace("PUBLIC_HEADER", "CONFIGURATIONS MyRelease\nPUBLIC_HEADER")
    c.save({"CMakeLists.txt": cmake})
    c.run("create . --name=hello --version=0.1 -s compiler.version=191 "
          "-s build_type=MyRelease -s:b build_type=MyRelease -tf=\"\"")

    # Prepare the actual consumer package
    c.save({"conanfile.py": conanfile,
            "CMakeLists.txt": cmakelist,
            "app.cpp": app})

    settings = {"compiler": "msvc",
                "compiler.version": "191",
                "compiler.runtime": "dynamic",
                "arch": "x86_64",
                "build_type": "MyRelease",
                }

    settings_h = " ".join('-s %s="%s"' % (k, v) for k, v in settings.items())
    settings_b = " ".join('-s:b %s="%s"' % (k, v) for k, v in settings.items())

    # Run the configure corresponding to this test case
    build_directory = os.path.join(c.current_folder, "build").replace("\\", "/")
    with c.chdir(build_directory):
        c.run("install .. %s %s -of=." % (settings_h, settings_b))
        toolchain = c.load("conan_toolchain.cmake")
        # As it is a custom configuration, the TRY_COMPILE_CONFIFURATION not defined
        assert "CMAKE_TRY_COMPILE_CONFIGURATION" not in toolchain
        assert os.path.isfile(os.path.join(c.current_folder,
                                           "hello-Target-myrelease.cmake"))

        c.run_command('cmake .. -G "Visual Studio 15" '
                      '-DCMAKE_TOOLCHAIN_FILE=conan_toolchain.cmake')
        c.run_command('cmake --build . --config MyRelease')
        c.run_command(r"MyRelease\\app.exe")
        assert "hello/0.1: Hello World Release!" in c.out
        assert "main: Release!" in c.out


@pytest.mark.tool("cmake")
def test_changing_build_type():
    client = TestClient(path_with_spaces=False)
    dep_conanfile = textwrap.dedent(r"""
       import os
       from conan import ConanFile
       from conan.tools.files import copy, save

       class Dep(ConanFile):
           settings = "build_type"
           def build(self):
               save(self, "hello.h",
               '# include <iostream>\n'
               'void hello(){{std::cout<<"BUILD_TYPE={}!!";}}'.format(self.settings.build_type))
           def package(self):
               copy(self, "*.h", self.source_folder, os.path.join(self.package_folder, "include"))
           """)
    client.save({"conanfile.py": dep_conanfile})
    client.run("create . --name=dep --version=0.1 -s build_type=Release")
    client.run("create . --name=dep --version=0.1 -s build_type=Debug")

    cmakelists = textwrap.dedent("""
        set(CMAKE_CXX_COMPILER_WORKS 1)
        set(CMAKE_CXX_ABI_COMPILED 1)
        cmake_minimum_required(VERSION 3.15)
        project(App CXX)

        # NOTE: NO MAP necessary!!!
        find_package(dep REQUIRED)
        add_executable(app app.cpp)
        target_link_libraries(app PRIVATE dep::dep)
        """)
    app = gen_function_cpp(name="main", includes=["hello"], calls=["hello"])
    pkg_conanfile = GenConanfile("pkg", "0.1").with_requires("dep/0.1").\
        with_generator("CMakeDeps").with_generator("CMakeToolchain").\
        with_settings("os", "compiler", "arch", "build_type")
    client.save({"conanfile.py": pkg_conanfile,
                 "CMakeLists.txt": cmakelists,
                 "app.cpp": app}, clean_first=True)

    # in MSVC multi-config -s pkg/*:build_type=Debug is not really necesary, toolchain do nothing
    # TODO: Challenge how to define consumer build_type for conanfile.txt
    client.run("install . -s pkg*:build_type=Debug -s build_type=Release")
    client.run_command("cmake . -DCMAKE_TOOLCHAIN_FILE=conan_toolchain.cmake "
                       "-DCMAKE_BUILD_TYPE=Debug")
    client.run_command("cmake --build . --config Debug")
    cmd = os.path.join(".", "Debug", "app") if platform.system() == "Windows" else "./app"
    client.run_command(cmd)
    assert "main: Debug!" in client.out
    assert "BUILD_TYPE=Release!!" in client.out



import os

from conan.internal.api.remotes.localdb import LocalDB
from conan.test.utils.test_files import temp_folder


def test_localdb():
    tmp_dir = temp_folder()
    db_file = os.path.join(tmp_dir, "dbfile")
    localdb = LocalDB(db_file)

    # Test write and read login
    user, token, access_token = localdb.get_login("myurl1")
    assert user is None
    assert token is None
    assert access_token is None

    localdb.store("pepe", "token", "access_token", "myurl1")
    user, token, access_token = localdb.get_login("myurl1")
    assert user == "pepe"
    assert token == "token"
    assert access_token == "access_token"
    assert "pepe" == localdb.get_username("myurl1")
